"""Import service types"""

from typing import Union

from typing_extensions import TypedDict


class AssetLike(TypedDict, total=False):
    """
    General type of an asset obejct through the import functions
    """

    content: Union[str, bytes]
    json_content: Union[dict, str, list]
    external_id: str
    status: str
    json_metadata: Union[str, dict]
    is_honeypot: bool
    id: str


class KiliResolverAsset(AssetLike, TypedDict, total=True):
    """
    Type of an asset object to be sent in Kili resolvers
    """

    content: Union[str, bytes]
    json_content: Union[dict, str, list]
    external_id: str
    status: str
    json_metadata: Union[str, dict]
    is_honeypot: bool
    id: str



"""
Types specific to import
"""
from typing import Dict, List, NewType, Optional

from pydantic import BaseModel, Extra, StrictInt, StrictStr, validator
from typing_extensions import Literal, Required, TypedDict

from kili.services.types import LabelType

Classes = NewType("Classes", Dict[int, str])
LabelFormat = Literal["yolo_v4", "yolo_v5", "yolo_v7", "kili", "raw"]


class LabelToImport(TypedDict, total=False):
    """
    Typed dictionary that specifies the keys and types for the labels to import.
    """

    author_id: str
    label_asset_external_id: Optional[str]
    label_asset_id: Optional[str]
    label_type: Optional[LabelType]
    seconds_to_label: Optional[int]
    path: Required[str]


class LabelData(BaseModel, extra=Extra.forbid):
    """
    Data about a label to append
    """

    asset_id: StrictStr
    json_response: dict
    author_id: Optional[StrictStr]
    seconds_to_label: Optional[StrictInt]
    model_name: Optional[StrictStr]


class _LabelsValidator(BaseModel, extra=Extra.forbid):
    """
    Validates the data about a label to append
    """

    labels: List[Dict]

    @validator("labels", each_item=True)
    def label_validator(cls, label):  # pylint: disable=no-self-argument
        """Validate the data of one label"""
        return LabelData(**label)



"""Project version queries."""

from typing import Generator, List, Optional, Union

from typeguard import typechecked

from kili.helpers import format_result, fragment_builder
from kili.queries.project_version.queries import (
    GQL_PROJECT_VERSION_COUNT,
    gql_project_version,
)
from kili.types import ProjectVersion as ProjectVersionType
from kili.utils.pagination import row_generator_from_paginated_calls


class QueriesProjectVersion:
    """Set of ProjectVersion queries."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @typechecked
    def project_version(
        self,
        project_id: str,
        first: Optional[int] = None,
        skip: int = 0,
        fields: List[str] = ["createdAt", "id", "content", "name", "projectId"],
        disable_tqdm: bool = False,
        as_generator: bool = False,
    ) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Get a generator or a list of project versions respecting a set of criteria.

        Args:
            project_id: Filter on Id of project
            fields: All the fields to request among the possible fields for the project versions
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#projectVersions) for all possible fields.
            first: Number of project versions to query
            skip: Number of project versions to skip (they are ordered by their date
                of creation, first to last).
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the project versions is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.
        """
        count_args = {"project_id": project_id}
        disable_tqdm = disable_tqdm or as_generator
        payload_query = {
            "where": {
                "projectId": project_id,
            },
        }
        project_versions_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_project_versions,
            count_args,
            self._query_project_versions,
            payload_query,
            fields,
            disable_tqdm,
        )

        if as_generator:
            return project_versions_generator
        return list(project_versions_generator)

    def _query_project_versions(self, skip: int, first: int, payload: dict, fields: List[str]):
        payload.update({"skip": skip, "first": first})
        _gql_project_version = gql_project_version(fragment_builder(fields, ProjectVersionType))
        result = self.auth.client.execute(_gql_project_version, payload)
        return format_result("data", result)

    @typechecked
    def count_project_versions(self, project_id: str) -> int:
        """Count the number of project versions.

        Args:
            project_id: Filter on ID of project

        Returns:
            The number of project versions with the parameters provided
        """
        variables = {
            "where": {"projectId": project_id},
        }
        result = self.auth.client.execute(GQL_PROJECT_VERSION_COUNT, variables)
        count = format_result("data", result)
        return int(count)



"""Service for exporting kili objects """

from pathlib import Path
from typing import Dict, List, Optional, Type

from typing_extensions import get_args

from kili.exceptions import NotFound
from kili.services.export.format.base import AbstractExporter, ExportParams
from kili.services.export.format.coco import CocoExporter
from kili.services.export.format.kili import KiliExporter
from kili.services.export.format.yolo import YoloExporter
from kili.services.export.logger import get_logger
from kili.services.export.repository import SDKContentRepository
from kili.services.export.types import ExportType, LabelFormat, SplitOption
from kili.services.types import LogLevel, ProjectId


def export_labels(  # pylint: disable=too-many-arguments, too-many-locals
    kili,
    asset_ids: Optional[List[str]],
    project_id: ProjectId,
    export_type: ExportType,
    label_format: LabelFormat,
    split_option: SplitOption,
    single_file: bool,
    output_file: str,
    disable_tqdm: bool,
    log_level: LogLevel,
) -> None:
    """
    Export the selected assets into the required format, and save it into a file archive.
    """
    if kili.count_projects(project_id=project_id) == 0:
        raise NotFound(f"project ID: {project_id}")

    export_params = ExportParams(
        assets_ids=asset_ids,
        project_id=project_id,
        export_type=export_type,
        label_format=label_format,
        split_option=split_option,
        single_file=single_file,
        output_file=Path(output_file),
    )

    logger = get_logger(log_level)

    content_repository = SDKContentRepository(
        kili.auth.api_endpoint,
        router_headers={
            "Authorization": f"X-API-Key: {kili.auth.api_key}",
        },
        verify_ssl=True,
    )

    if label_format in get_args(LabelFormat):
        format_exporter_selector_mapping: Dict[str, Type[AbstractExporter]] = {
            "raw": KiliExporter,
            "kili": KiliExporter,
            "coco": CocoExporter,
            "yolo_v4": YoloExporter,
            "yolo_v5": YoloExporter,
            "yolo_v7": YoloExporter,
        }
        assert set(format_exporter_selector_mapping.keys()) == set(
            get_args(LabelFormat)
        )  # ensures full mapping
        exporter_class = format_exporter_selector_mapping[label_format]
        exporter_class(
            export_params, kili, logger, disable_tqdm, content_repository
        ).export_project()
    else:
        raise ValueError(f'Label format "{label_format}" is not implemented or does not exist.')



"""
Python SDK service layer
"""

from kili.services.asset_import import import_assets
from kili.services.export import export_labels
from kili.services.label_import import import_labels_from_files

__all__ = ["export_labels", "import_assets", "import_labels_from_files"]



"""
Queries of issue queries
"""


def gql_issues(fragment):
    """
    Return the GraphQL issues query
    """
    return f"""
query ($where: IssueWhere!, $first: PageSize!, $skip: Int!) {{
  data: issues(where: $where, first: $first, skip: $skip) {{
    {fragment}
  }}
}}
"""


GQL_ISSUES_COUNT = """
query($where: IssueWhere!) {
  data: countIssues(where: $where)
}
"""



"""Project mutations."""

from json import dumps
from typing import Dict, Optional, Union

from typeguard import typechecked

from ...helpers import format_result
from .helpers import verify_argument_ranges
from .queries import (
    GQL_APPEND_TO_ROLES,
    GQL_CREATE_PROJECT,
    GQL_DELETE_FROM_ROLES,
    GQL_DELETE_PROJECT,
    GQL_GQL_UPDATE_PROPERTIES_IN_PROJECT_USER,
    GQL_PROJECT_DELETE_ASYNCHRONOUSLY,
    GQL_UPDATE_PROPERTIES_IN_PROJECT,
    GQL_UPDATE_PROPERTIES_IN_ROLE,
)


class MutationsProject:
    """Set of Project mutations."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @typechecked
    def append_to_roles(
        self, project_id: str, user_email: str, role: str = "LABELER"
    ) -> Dict[str, Union[str, dict, list, None]]:
        """Add a user to a project.

        !!! info
            If the user does not exist in your organization, he/she is invited and added
                both to your organization and project. This function can also be used to change
                the role of the user in the project.

        Args:
            project_id: Identifier of the project
            user_email: The email of the user.
                This email is used as the unique identifier of the user.
            role: One of {"ADMIN", "TEAM_MANAGER", "REVIEWER", "LABELER"}.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.


        Examples:
            >>> kili.append_to_roles(project_id=project_id, user_email='john@doe.com')
        """
        variables = {
            "data": {"role": role, "userEmail": user_email},
            "where": {"id": project_id},
        }
        result = self.auth.client.execute(GQL_APPEND_TO_ROLES, variables)
        return format_result("data", result)

    @typechecked
    def update_properties_in_project(
        self,
        project_id: str,
        can_navigate_between_assets: Optional[bool] = None,
        can_skip_asset: Optional[bool] = None,
        consensus_mark: Optional[float] = None,
        consensus_tot_coverage: Optional[int] = None,
        description: Optional[str] = None,
        honeypot_mark: Optional[float] = None,
        instructions: Optional[str] = None,
        input_type: Optional[str] = None,
        json_interface: Optional[dict] = None,
        min_consensus_size: Optional[int] = None,
        number_of_assets: Optional[int] = None,
        number_of_skipped_assets: Optional[int] = None,
        number_of_remaining_assets: Optional[int] = None,
        number_of_reviewed_assets: Optional[int] = None,
        review_coverage: Optional[int] = None,
        should_relaunch_kpi_computation: Optional[bool] = None,
        title: Optional[str] = None,
        use_honeypot: Optional[bool] = None,
        metadata_types: Optional[dict] = None,
    ) -> Dict[str, str]:
        """Update properties of a project.

        Args:
            project_id: Identifier of the project.
            can_navigate_between_assets:
                Activate / Deactivate the use of next and previous buttons in labeling interface.
            can_skip_asset: Activate / Deactivate the use of skip button in labeling interface.
            consensus_mark: Should be between 0 and 1.
            consensus_tot_coverage: Should be between 0 and 100.
                It is the percentage of the dataset that will be annotated several times.
            description : Description of the project.
            honeypot_mark : Should be between 0 and 1
            instructions : Instructions of the project.
            input_type: Currently, one of `AUDIO`, `IMAGE`, `PDF`, `TEXT`,
                `VIDEO`, `VIDEO_LEGACY`.
            json_interface: The json parameters of the project, see Edit your interface.
            min_consensus_size: Should be between 1 and 10
                Number of people that will annotate the same asset, for consensus computation.
            number_of_assets: Defaults to 0
            number_of_skipped_assets: Defaults to 0
            number_of_remaining_assets: Defaults to 0
            number_of_reviewed_assets: Defaults to 0
            review_coverage: Allow to set the percentage of assets
                that will be queued in the review interface.
                Should be between 0 and 100
            should_relaunch_kpi_computation: Technical field, added to indicate changes
                in honeypot or consensus settings
            title: Title of the project
            use_honeypot: Activate / Deactivate the use of honeypot in the project
            metadata_types: Types of the project metadata.
                Should be a `dict` of metadata fields name as keys and metadata types as values.
                Currently, possible types are: `string`, `number`

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.update_properties_in_project(project_id=project_id, title='New title')

        !!! example "Change Metadata Types"
            Metadata fields are by default interpreted as `string` types. To change the type
            of a metadata field, you can use the `update_properties_in_project` function with the
            metadata_types argument. `metadata_types` is given as a dict of metadata field names
            as keys and metadata types as values.
            Example:
            ```
            kili.update_properties_in_project(
                project_id = project_id,
                metadata_types = {
                    'customConsensus': 'number',
                    'sensitiveData': 'string',
                    'uploadedFromCloud': 'string',
                    'modelLabelErrorScore': 'number'
                }
            )
            ```
            Not providing a type for a metadata field or providing an unsupported one
            will default to the `string` type.
        """
        verify_argument_ranges(consensus_tot_coverage, min_consensus_size, review_coverage)

        variables = {
            "canNavigateBetweenAssets": can_navigate_between_assets,
            "canSkipAsset": can_skip_asset,
            "consensusMark": consensus_mark,
            "consensusTotCoverage": consensus_tot_coverage,
            "description": description,
            "honeypotMark": honeypot_mark,
            "instructions": instructions,
            "inputType": input_type,
            "jsonInterface": dumps(json_interface) if json_interface is not None else None,
            "metadataTypes": metadata_types,
            "minConsensusSize": min_consensus_size,
            "numberOfAssets": number_of_assets,
            "numberOfSkippedAssets": number_of_skipped_assets,
            "numberOfRemainingAssets": number_of_remaining_assets,
            "numberOfReviewedAssets": number_of_reviewed_assets,
            "projectID": project_id,
            "reviewCoverage": review_coverage,
            "shouldRelaunchKpiComputation": should_relaunch_kpi_computation,
            "title": title,
            "useHoneyPot": use_honeypot,
        }
        result = self.auth.client.execute(GQL_UPDATE_PROPERTIES_IN_PROJECT, variables)
        return format_result("data", result)

    @typechecked
    def create_project(
        self,
        input_type: str,
        json_interface: dict,
        title: str,
        description: str = "",
        project_type: Optional[str] = None,
    ) -> Dict:
        # pylint: disable=line-too-long
        """Create a project.

        Args:
            input_type : Currently, one of {AUDIO, IMAGE, PDF, TEXT, URL, VIDEO, VIDEO_LEGACY, NA}
            json_interface: The json parameters of the project, see Edit your interface.
            title : Title of the project
            description : Description of the project
            project_type:
                Currently, one of {
                    `IMAGE_CLASSIFICATION_SINGLE`,
                    `IMAGE_CLASSIFICATION_MULTI`,
                    `IMAGE_OBJECT_DETECTION_RECTANGLE`,
                    `IMAGE_OBJECT_DETECTION_POLYGON`,
                    `IMAGE_OBJECT_DETECTION_SEMANTIC`,
                    `OCR, PDF_CLASSIFICATION_SINGLE`,
                    `PDF_CLASSIFICATION_MULTI`,
                    `TEXT_CLASSIFICATION_SINGLE`,
                    `TEXT_CLASSIFICATION_MULTI`,
                    `TEXT_TRANSCRIPTION, TEXT_NER`,
                    `VIDEO_CLASSIFICATION_SINGLE`,
                    `VIDEO_FRAME_CLASSIFICATION`,
                    `VIDEO_FRAME_OBJECT_TRACKING`,
                    `SPEECH_TO_TEXT`
                }

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.create_project(input_type='IMAGE', json_interface=json_interface, title='Example')

        !!! example "Recipe"
            For more detailed examples on how to create projects,
                see [the recipe](https://github.com/kili-technology/kili-python-sdk/blob/master/recipes/create_project.ipynb).
        """
        variables = {
            "data": {
                "description": description,
                "inputType": input_type,
                "jsonInterface": dumps(json_interface),
                "projectType": project_type,
                "title": title,
            }
        }
        result = self.auth.client.execute(GQL_CREATE_PROJECT, variables)
        return format_result("data", result)

    @typechecked
    def update_properties_in_role(self, role_id: str, project_id: str, user_id: str, role: str):
        """Update properties of a role.

        !!! info
            To be able to change someone's role, you must be either of:

            - an admin of the project
            - a team manager of the project
            - an admin of the organization

        Args:
            role_id: Role identifier of the user. E.g. : 'to-be-deactivated'
            project_id: Identifier of the project
            user_id: The email or identifier of the user with updated role
            role: The new role.
                Possible choices are: `ADMIN`, `TEAM_MANAGER`, `REVIEWER`, `LABELER`

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {
            "roleID": role_id,
            "projectID": project_id,
            "userID": user_id,
            "role": role,
        }
        result = self.auth.client.execute(GQL_UPDATE_PROPERTIES_IN_ROLE, variables)
        return format_result("data", result)

    @typechecked
    def delete_from_roles(self, role_id: str):
        """Delete users by their role_id.

        Args:
            role_id : Identifier of the project user (not the ID of the user)

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {"where": {"id": role_id}}
        result = self.auth.client.execute(GQL_DELETE_FROM_ROLES, variables)
        return format_result("data", result)

    @typechecked
    def update_properties_in_project_user(
        self,
        project_user_id: str,
        consensus_mark: Optional[float] = None,
        honeypot_mark: Optional[float] = None,
        number_of_labeled_assets: Optional[int] = None,
        starred: Optional[bool] = None,
        total_duration: Optional[int] = None,
    ):
        """
        Update properties of a project-user tuple

        Args:
            project_user_id : Identifier of the project user
            consensus_mark: Should be between 0 and 1.
            honeypot_mark: Should be between 0 and 1.
            number_of_labeled_assets: Number of assets the user labeled in the project.
            starred: Whether to star the project in the project list.
            total_duration: Total time the user spent in the project.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> for project_user in project_users:
            ...     kili.update_properties_in_project_user(
                        project_user_id=project_user['id'],
                        honeypot_mark=0)
        """
        variables = {
            "consensusMark": consensus_mark,
            "honeypotMark": honeypot_mark,
            "numberOfLabeledAssets": number_of_labeled_assets,
            "projectUserID": project_user_id,
            "starred": starred,
            "totalDuration": total_duration,
        }
        result = self.auth.client.execute(GQL_GQL_UPDATE_PROPERTIES_IN_PROJECT_USER, variables)
        return format_result("data", result)

    @typechecked
    def internal_delete_project(self, project_id: str):
        """Delete project permanently.
        WARNING: This resolver is for internal use by Kili Technology only.

        Args:
            project_id: Identifier of the project

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {"projectID": project_id}
        result = self.auth.client.execute(GQL_DELETE_PROJECT, variables)
        return format_result("data", result)

    @typechecked
    def delete_project(self, project_id: str) -> str:
        """
        Delete a project permanently.

        Args:
            project_id: Identifier of the project

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {"where": {"id": project_id}}
        result = self.auth.client.execute(GQL_PROJECT_DELETE_ASYNCHRONOUSLY, variables)
        return format_result("data", result)

    @typechecked
    def archive_project(self, project_id: str):
        """
        Archive a project.

        Args:
            project_id: Identifier of the project

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """

        variables = {
            "projectID": project_id,
            "archived": True,
        }

        result = self.auth.client.execute(GQL_UPDATE_PROPERTIES_IN_PROJECT, variables)
        return format_result("data", result)

    @typechecked
    def unarchive_project(self, project_id: str):
        """
        Unarchive a project.

        Args:
            project_id: Identifier of the project

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """

        variables = {
            "projectID": project_id,
            "archived": False,
        }

        result = self.auth.client.execute(GQL_UPDATE_PROPERTIES_IN_PROJECT, variables)
        return format_result("data", result)



"""
Label Importers
"""
import csv
import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Type, cast

import yaml

from kili.helpers import get_file_paths_to_upload
from kili.services.helpers import (
    check_exclusive_options,
    get_external_id_from_file_path,
)
from kili.services.label_import.exceptions import (
    LabelParsingError,
    MissingExternalIdError,
    MissingMetadataError,
    MissingTargetJobError,
)
from kili.services.label_import.parser import (
    AbstractLabelParser,
    KiliRawLabelParser,
    YoloLabelParser,
)
from kili.services.label_import.types import Classes, LabelFormat, LabelToImport
from kili.services.types import LabelType, LogLevel, ProjectId
from kili.utils.tqdm import tqdm


class LoggerParams(NamedTuple):
    """
    Contains all parameters related to logging.
    """

    disable_tqdm: bool
    level: LogLevel


class AbstractLabelImporter(ABC):
    """
    Abstract Label Importer
    """

    def __init__(self, kili, logger_params: LoggerParams, input_format: LabelFormat):
        self.kili = kili
        self.logger_params = logger_params
        self.input_format: LabelFormat = input_format

        self.logger = logging.getLogger("kili.services.label_import")
        self.logger.setLevel(logger_params.level)

    def process_from_files(  # pylint: disable=too-many-arguments
        self,
        labels_file_path: Optional[Path],
        labels_files: Optional[List[Path]],
        meta_file_path: Optional[Path],
        project_id: ProjectId,
        target_job_name: Optional[str],
        model_name: Optional[str],
        is_prediction: bool,
    ):
        """
        Performs the import from the label files
        """
        self._check_arguments_compatibility(meta_file_path, target_job_name)
        class_by_id = self._read_classes_from_meta_file(meta_file_path, self.input_format)
        label_parser_class = self._select_label_parser()
        label_parser = label_parser_class(class_by_id, target_job_name)

        self.logger.warning("Importing labels")
        labels = self.extract_from_files(labels_file_path, labels_files)
        if is_prediction:
            assert model_name
            self._import_as_predictions(labels, label_parser, project_id, model_name)
        else:
            self._import_as_labels(labels, label_parser, project_id)

        self.logger.warning(print(f"{len(labels)} labels have been successfully imported"))

    def _import_as_labels(
        self, labels: List[LabelToImport], label_parser: AbstractLabelParser, project_id: ProjectId
    ):
        for label in tqdm(labels, disable=self.logger_params.disable_tqdm):
            kwargs = dict(label.copy())
            del kwargs["path"]
            try:
                json_response = label_parser.parse(Path(label["path"]))
            except Exception as exc:
                path = label["path"]
                raise LabelParsingError(f"Failed to parse the file {path}") from exc

            self.kili.append_to_labels(
                json_response=json_response,
                project_id=project_id,
                **kwargs,
            )

    def _import_as_predictions(
        self,
        labels: List[LabelToImport],
        label_parser: AbstractLabelParser,
        project_id: ProjectId,
        model_name: str,
    ):
        assert model_name
        json_response_array: List[Dict] = []
        external_id_array: List[str] = []
        model_name_array: List[str] = []
        for label in tqdm(labels, disable=self.logger_params.disable_tqdm):
            kwargs = dict(label.copy())
            del kwargs["path"]
            json_response_array.append(label_parser.parse(Path(label["path"])))
            external_id = label.get("label_asset_external_id")
            if external_id is None:
                path = label["path"]
                raise MissingExternalIdError(f"There is no associated asset external id to {path}")
            external_id_array.append(external_id)
            model_name_array.append(model_name)
        self.kili.create_predictions(
            json_response_array=json_response_array,
            external_id_array=external_id_array,
            model_name_array=model_name_array,
            project_id=project_id,
        )

    @staticmethod
    @abstractmethod
    def _check_arguments_compatibility(
        meta_file_path: Optional[Path], target_job_name: Optional[str]
    ):
        pass

    @classmethod
    @abstractmethod
    def _read_classes_from_meta_file(
        cls, meta_file_path: Optional[Path], input_format: LabelFormat
    ) -> Optional[Classes]:
        pass

    @abstractmethod
    def _select_label_parser(self) -> Type[AbstractLabelParser]:
        pass

    @abstractmethod
    def _get_label_file_extension(self) -> str:
        pass

    def extract_from_files(
        self, labels_file_path: Optional[Path], labels_files: Optional[List[Path]]
    ) -> List[LabelToImport]:
        """
        Extracts the labels files and their metadata from the label files (given
        explicitely or through a CSV)
        """
        check_exclusive_options(labels_file_path, labels_files)

        labels = []
        if labels_files is not None and len(labels_files) > 0:
            label_paths = get_file_paths_to_upload(
                [str(f) for f in labels_files],
                lambda path: path.endswith(self._get_label_file_extension()),
                verbose=self.logger_params.level in ["INFO", "DEBUG"],
            )
            if len(label_paths) == 0:
                raise ValueError(
                    "No label files to upload. "
                    "Check that the paths exist and file types are .json"
                )
            external_ids = [get_external_id_from_file_path(Path(path)) for path in label_paths]
            labels = [
                LabelToImport(path=p, label_asset_external_id=e_id)
                for (p, e_id) in zip(label_paths, external_ids)
            ]

        elif labels_file_path is not None:
            labels = self._read_labels_file_path(labels_file_path)

            if len(labels) == 0:
                raise ValueError(f"No label file was found in csv: {labels_file_path}")
        return labels

    @classmethod
    def _read_labels_file_path(cls, labels_file: Path) -> List[LabelToImport]:
        """
        Read CSV and fill a list of labels to import.
        """
        data = []
        with labels_file.open("r") as l_f:
            csv_reader = csv.reader(l_f)
            data = []
            headers = []
            for ind, row in enumerate(csv_reader):
                if ind == 0:
                    headers = row
                else:
                    headers_row = zip(headers, row)
                    str_dict = dict(headers_row)

                    typed_dict: Dict[str, Any] = {
                        "path": str_dict["path"],
                    }
                    if len(str_dict.get("label_asset_id", "")):
                        typed_dict["label_asset_id"] = str_dict["label_asset_id"]

                    if len(str_dict.get("label_asset_external_id", "")):
                        typed_dict["label_asset_external_id"] = str_dict["label_asset_external_id"]

                    if len(str_dict.get("label_type", "")):
                        typed_dict["label_type"] = cast(LabelType, str_dict["label_type"])

                    if len(str_dict.get("seconds_to_label", "")):
                        typed_dict["seconds_to_label"] = int(str_dict["seconds_to_label"])

                    if len(str_dict.get("author_id", "")):
                        typed_dict["author_id"] = str_dict["author_id"]

                    data.append(LabelToImport(**typed_dict))
        return data


class YoloLabelImporter(AbstractLabelImporter):
    """
    Label importer in the yolo format.
    """

    def _get_label_file_extension(self) -> str:
        return ".txt"

    @staticmethod
    def _check_arguments_compatibility(
        meta_file_path: Optional[Path], target_job_name: Optional[str]
    ):
        if meta_file_path is None:
            raise MissingMetadataError("Meta file is needed to import the label")
        if target_job_name is None:
            raise MissingTargetJobError("A target job name is needed to import the label")

    @staticmethod
    def _read_classes_from_meta_file(
        meta_file_path: Optional[Path], input_format: LabelFormat
    ) -> Classes:
        assert meta_file_path
        classes: Classes = Classes({})
        if input_format == "yolo_v4":
            try:
                # layout: id class\n
                with meta_file_path.open("r", encoding="utf-8") as m_f:
                    csv_reader = csv.reader(m_f, delimiter=" ")
                    classes = Classes({int(r[0]): r[1] for r in csv_reader if r[0] != " "})
            except (ValueError, IndexError):
                with meta_file_path.open("r", encoding="utf-8") as m_f:
                    classes = Classes(dict(enumerate(m_f.read().splitlines())))

        elif input_format == "yolo_v5":
            with meta_file_path.open("r", encoding="utf-8") as m_f:
                m_d = yaml.load(m_f, yaml.FullLoader)
                classes = Classes(m_d["names"])
        elif input_format == "yolo_v7":
            with meta_file_path.open("r", encoding="utf-8") as m_f:
                m_d = yaml.load(m_f, yaml.FullLoader)
                classes = Classes(dict(enumerate(m_d["names"])))
        else:
            raise NotImplementedError(f"The format f{input_format} does not have a metadata parser")

        return classes

    def _select_label_parser(self) -> Type[AbstractLabelParser]:
        return YoloLabelParser


class KiliRawLabelImporter(AbstractLabelImporter):
    """
    # Label importer in the Kili format.
    """

    def _get_label_file_extension(self) -> str:
        return ".json"

    @staticmethod
    def _check_arguments_compatibility(
        meta_file_path: Optional[Path], target_job_name: Optional[str]
    ):
        pass

    @classmethod
    def _read_classes_from_meta_file(
        cls, meta_file_path: Optional[Path], input_format: LabelFormat
    ) -> Optional[Classes]:
        return None

    def _select_label_parser(self) -> Type[AbstractLabelParser]:
        return KiliRawLabelParser



"""
Queries of project queries
"""


def gql_projects(fragment: str):
    """
    Return the GraphQL projects query
    """
    return f"""
query($where: ProjectWhere!, $first: PageSize!, $skip: Int!) {{
  data: projects(where: $where, first: $first, skip: $skip) {{
    {fragment}
  }}
}}
"""


GQL_PROJECTS_COUNT = """
query($where: ProjectWhere!) {
  data: countProjects(where: $where)
}
"""



"""
Set of common functions used by different export formats
"""
from typing import List, Optional

from kili.services.export.types import ExportType

DEFAULT_FIELDS = [
    "id",
    "content",
    "externalId",
    "jsonMetadata",
    "labels.author.id",
    "labels.author.email",
    "labels.author.firstname",
    "labels.author.lastname",
    "labels.jsonResponse",
    "labels.createdAt",
    "labels.isLatestLabelForUser",
    "labels.labelType",
    "labels.modelName",
]
LATEST_LABEL_FIELDS = [
    "id",
    "content",
    "externalId",
    "jsonContent",
    "jsonMetadata",
    "latestLabel.author.id",
    "latestLabel.author.email",
    "latestLabel.jsonResponse",
    "latestLabel.author.firstname",
    "latestLabel.author.lastname",
    "latestLabel.createdAt",
    "latestLabel.isLatestLabelForUser",
    "latestLabel.labelType",
    "latestLabel.modelName",
]


def attach_name_to_assets_labels_author(assets, export_type: ExportType):
    """
    Adds `name` field for author, by concatenating his/her first and last name
    """
    for asset in assets:
        if export_type == "latest":
            latest_label = asset["latestLabel"]
            if latest_label:
                firstname = latest_label["author"]["firstname"]
                lastname = latest_label["author"]["lastname"]
                latest_label["author"]["name"] = f"{firstname} {lastname}"
            continue
        for label in asset.get("labels", []):
            firstname = label["author"]["firstname"]
            lastname = label["author"]["lastname"]
            label["author"]["name"] = f"{firstname} {lastname}"


def fetch_assets(  # pylint: disable=too-many-arguments
    kili,
    project_id: str,
    asset_ids: Optional[List[str]],
    export_type,
    label_type_in=None,
    disable_tqdm: bool = False,
    download_media: bool = False,
):
    """
    Fetches assets where ID are in asset_ids if the list has more than one element,
    else all the assets of the project. If download media is passed, the media are
    downloaded into the `$HOME/.cache` folder.

    Parameters
    ----------
    - project_id: project id
    - assets_ids: list of asset IDs
    - export_type: type of export (latest label or all labels)
    - label_type_in: types of label to fetch (default, reviewed, ...)
    - disable_tqdm: tell to disable tqdm
    - download_media: tell to download the media in the cache folder.
    """
    fields = get_fields_to_fetch(export_type)
    assets = None

    if asset_ids is not None and len(asset_ids) > 0:
        assets = kili.assets(
            asset_id_in=asset_ids,
            project_id=project_id,
            fields=fields,
            label_type_in=label_type_in,
            disable_tqdm=disable_tqdm,
            download_media=download_media,
        )
    else:
        assets = kili.assets(
            project_id=project_id,
            fields=fields,
            label_type_in=label_type_in,
            disable_tqdm=disable_tqdm,
            download_media=download_media,
        )
    attach_name_to_assets_labels_author(assets, export_type)
    return assets


def get_fields_to_fetch(export_type: ExportType):
    """
    Returns the fields to fetch depending on the export type
    """
    if export_type == "latest":
        return LATEST_LABEL_FIELDS
    return DEFAULT_FIELDS



"""
Functions to import assets into a TEXT project
"""
import os
from enum import Enum
from typing import List, Optional, Tuple

from kili.helpers import is_url

from .base import (
    BaseAssetImporter,
    BatchParams,
    ContentBatchImporter,
    JsonContentBatchImporter,
)
from .exceptions import ImportValidationError
from .types import AssetLike


class TextDataType(Enum):
    """
    Text data type
    """

    RICH_TEXT = "RICH_TEXT"
    LOCAL_FILE = "LOCAL_FILE"
    HOSTED_FILE = "HOSTED_FILE"
    RAW_TEXT = "RAW_TEXT"


class RawTextBatchImporter(ContentBatchImporter):
    """
    class for importing a batch of raw text assets into a TEXT project
    """

    def get_content_type_and_data_from_content(self, content: Optional[str]) -> Tuple[str, str]:
        """
        Returns the data of the content (path) and its content type
        """
        return content or "", "text/plain"


class TextDataImporter(BaseAssetImporter):
    """
    Class for importing data into a TEXT project
    """

    @staticmethod
    def get_data_type(assets: List[AssetLike]) -> TextDataType:
        """
        Determine the type of data to upload from the service payload
        """
        content_array = [asset.get("content", "") for asset in assets]
        has_local_file = any(os.path.exists(content) for content in content_array)
        has_hosted_file = any(is_url(content) for content in content_array)
        has_json_content = any(asset.get("json_content") for asset in assets)
        if has_json_content:
            if any(content_array):
                raise ImportValidationError(
                    "Cannot import content when importing a Rich Text asset"
                )
            return TextDataType.RICH_TEXT
        if has_local_file and has_hosted_file:
            raise ImportValidationError(
                """
                Cannot upload hosted data and local files at the same time.
                Please separate the assets into 2 calls
                """
            )
        if has_local_file:
            return TextDataType.LOCAL_FILE
        if has_hosted_file:
            return TextDataType.HOSTED_FILE
        return TextDataType.RAW_TEXT

    def import_assets(self, assets: List[AssetLike]):
        """
        Import TEXT assets into Kili.
        """
        data_type = self.get_data_type(assets)
        assets = self.filter_duplicate_external_ids(assets)
        if data_type == TextDataType.LOCAL_FILE:
            assets = self.filter_local_assets(assets, self.raise_error)
            batch_params = BatchParams(is_hosted=False, is_asynchronous=False)
            batch_importer = ContentBatchImporter(
                self.auth, self.project_params, batch_params, self.pbar
            )
        elif data_type == TextDataType.HOSTED_FILE:
            batch_params = BatchParams(is_hosted=True, is_asynchronous=False)
            batch_importer = ContentBatchImporter(
                self.auth, self.project_params, batch_params, self.pbar
            )
        elif data_type == TextDataType.RAW_TEXT:
            for asset in assets:
                if "content" in asset and isinstance(asset["content"], str):
                    asset["content"] = asset["content"].encode("utf-8")

            batch_params = BatchParams(is_hosted=False, is_asynchronous=False)
            batch_importer = RawTextBatchImporter(
                self.auth, self.project_params, batch_params, self.pbar
            )
        elif data_type == TextDataType.RICH_TEXT:
            batch_params = BatchParams(is_hosted=False, is_asynchronous=False)
            batch_importer = JsonContentBatchImporter(
                self.auth, self.project_params, batch_params, self.pbar
            )
        else:
            raise ImportValidationError
        result = self.import_assets_by_batch(assets, batch_importer)
        return result



"""
Set of common functions used by plugins
"""

DUPLICATE_ERROR_SUBSTRING = "an entity Plugin already exists with value"


def check_errors_plugin_upload(result, file_path, plugin_name):
    """
    Check if the error during upload is because the plugin already
    exists and print the upload function for the user
    """
    if "errors" in result and DUPLICATE_ERROR_SUBSTRING in result["errors"][0]["message"]:
        print(
            "Hint: A plugin with this name already exist, "
            "if you want to override it you can use the command "
            f'kili.update_plugin(file_path="{file_path}", plugin_name="{plugin_name}")'
        )



"""
Helpers for the label mutations
"""
import json
from os import PathLike
from typing import Dict, List


def generate_create_predictions_arguments(
    label_paths: List[PathLike],
    external_id_array: List[str],
    model_name: str,
    project_id: str,
) -> Dict:
    """
    Generate the arguments of create prediction mutation given
    a list of external ids and paths to json response files

    Args:
        label_paths: a list of paths to json files storing label responses
        external_id_array: a list of external ids
        model_name: the model name
        project_id: Project ID
    """
    json_response_array = []
    for path in label_paths:
        with open(path, encoding="utf-8") as label_file:
            json_response = json.load(label_file)
            json_response_array.append(json_response)
    return {
        "project_id": project_id,
        "json_response_array": json_response_array,
        "model_name_array": [model_name] * len(external_id_array),
        "external_id_array": external_id_array,
    }



"""
    CLI's project list subcommand
"""

from typing import Dict, List, Optional, cast

import click
import numpy as np
import pandas as pd
from tabulate import tabulate

from kili.cli.common_args import Options
from kili.cli.helpers import get_kili_client


@click.command(name="list")
@Options.api_key
@Options.endpoint
@Options.tablefmt
@click.option(
    "--max",
    "first",
    type=int,
    help="Maximum number of project to display.",
    default=100,
)
def list_projects(api_key: Optional[str], endpoint: Optional[str], tablefmt: str, first: int):
    """
    List your projects

    \b
    !!! Examples
        ```
        kili project list --max 10 --stdout-format pretty
        ```

    """
    kili = get_kili_client(api_key=api_key, api_endpoint=endpoint)
    projects = cast(
        List[Dict],
        kili.projects(
            fields=[
                "title",
                "id",
                "description",
                "numberOfAssets",
                "numberOfRemainingAssets",
                "numberOfReviewedAssets",
            ],
            first=first,
            disable_tqdm=True,
        ),
    )
    projects = pd.DataFrame(projects)
    projects["progress"] = projects.apply(
        lambda x: round((1 - x["numberOfRemainingAssets"] / x["numberOfAssets"]) * 100, 1)
        if x["numberOfAssets"] != 0
        else np.nan,
        axis=1,
    )

    # Add '%' to PROGRESS if PROGRESS is not nan
    projects["PROGRESS"] = [
        (str(progress) + "%") if progress >= 0 else progress for progress in projects["progress"]
    ]
    # If description or title has more than 50 characters, truncate after 47 and add '...'
    projects["DESCRIPTION"] = [
        (description[:47] + "...").replace("\n", "") if len(description) > 50 else description
        for description in projects["description"]
    ]
    projects["TITLE"] = [
        (title[:47] + "...") if len(title) > 50 else title for title in projects["title"]
    ]
    projects["ID"] = projects["id"]

    projects = projects[["TITLE", "ID", "PROGRESS", "DESCRIPTION"]]
    print(
        tabulate(
            projects,
            headers="keys",
            tablefmt=tablefmt,
            showindex=False,
            colalign=("left", "left", "right", "left"),
        )
    )



"""API authentication module"""
import os
import warnings
from datetime import datetime, timedelta

import requests
from requests import adapters
from urllib3 import Retry

from kili import __version__
from kili.graphql_client import GraphQLClient, GraphQLClientName
from kili.helpers import format_result
from kili.queries.api_key import QueriesApiKey
from kili.queries.user.queries import GQL_ME
from kili.types import User

MAX_RETRIES = 20

warnings.filterwarnings("default", module="kili", category=DeprecationWarning)


def get_version_without_patch(version):
    """Return the version of Kili API removing the patch version.

    Args:
        version
    """
    return ".".join(version.split(".")[:-1])


class KiliAuth:  # pylint: disable=too-many-instance-attributes
    """
    from kili.client import Kili
    kili = Kili(api_key=api_key)
    assets = kili.assets(project_id=project_id)
    """

    def __init__(
        self, api_key: str, api_endpoint: str, client_name: GraphQLClientName, verify=True
    ):
        self.session = requests.Session()
        self.client_name = client_name
        self.api_key = api_key
        self.api_endpoint = api_endpoint

        self.verify = verify

        try:
            self.check_versions_match(api_endpoint)
        except:  # pylint: disable=bare-except
            message = (
                "We could not check the version, there might be a version"
                "mismatch or the app might be in deployment"
            )
            warnings.warn(message, UserWarning)

        adapter = requests.adapters.HTTPAdapter(max_retries=MAX_RETRIES)  # type: ignore
        self.session.mount("https://", adapter)
        self.session.mount("http://", adapter)
        self.client = GraphQLClient(
            api_endpoint,
            client_name,
            self.session,
            verify=self.verify,
        )
        self.client.inject_token("X-API-Key: " + api_key)

        user = self.get_user()

        if user is None or user["id"] is None or user["email"] is None:
            raise Exception("No user attached to the API key was found")

        self.user_id = user["id"]
        self.user_email = user["email"]
        self.check_expiry_of_key_is_close(api_key)

    @staticmethod
    def _get_retry_policy() -> adapters.Retry:
        number_of_trials = None
        try:
            number_of_trials = int(os.getenv("KILI_SDK_TRIALS_NUMBER", "10"))
        except ValueError:
            number_of_trials = 10
        retry_policy = Retry(connect=number_of_trials, read=0, redirect=0, status=0, other=0)
        return retry_policy

    def __del__(self):
        self.session.close()

    def check_versions_match(self, api_endpoint):
        """Check that the versions of Kili Python SDK and Kili API are the same

        Args:
            api_endpoint: url of the Kili API
        """
        url = api_endpoint.replace("/graphql", "/version")
        response = requests.get(url, verify=self.verify, timeout=30).json()
        version = response["version"]
        if get_version_without_patch(version) != get_version_without_patch(__version__):
            message = (
                "Kili Python SDK version should match with Kili API version.\n"
                + f'Please install version: "pip install kili=={version}"'
            )
            warnings.warn(message, UserWarning)

    def check_expiry_of_key_is_close(self, api_key):
        """Check that the expiration date of the api_key is not too close.

        Args:
            api_key: key used to connect to the Kili API
        """
        duration_days = 365
        warn_days = 30
        queries = QueriesApiKey(self)
        key_object = queries.api_keys(api_key=api_key, fields=["createdAt"], disable_tqdm=True)
        key_creation = datetime.strptime(list(key_object)[0]["createdAt"], "%Y-%m-%dT%H:%M:%S.%fZ")
        key_expiry = key_creation + timedelta(days=duration_days)
        key_remaining_time = key_expiry - datetime.now()
        key_soon_deprecated = key_remaining_time < timedelta(days=warn_days)
        if key_soon_deprecated:
            message = f"""
Your api key will be deprecated on {key_expiry:%Y-%m-%d}.
You should generate a new one on My account > API KEY."""
            warnings.warn(message, UserWarning)

    def get_user(self) -> User:
        """Get the current user from the api_key provided"""
        result = self.client.execute(GQL_ME)
        return format_result("data", result)



"""Develop Plugins for Kili"""

import logging
from typing import Optional

from kili.client import Kili
from kili.types import Label


class PluginCore:
    """
    Kili Plugin core class

    :param kili: kili instance that plugins can make use of
    :param project_id: the project on which plugin is ran

    Implements
    on_submit(self, label: Label, asset_id: str, project_id: str)
    on_review(self, label: Label, asset_id: str, project_id: str)

    # Warning : if using a custom init, be sure to call super().__init__()
    """

    logger: logging.Logger
    kili: Kili
    project_id: str

    def __init__(
        self, kili: Kili, project_id: str, logger: Optional[logging.Logger] = None
    ) -> None:
        self.kili = kili
        self.project_id = project_id
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger()

    def on_submit(
        self,
        label: Label,
        asset_id: str,
    ) -> None:
        """
        Handler for the submit action, triggered when a default label is submitted into Kili
        :param label: label submitted to Kili
        :param asset_id: id of the asset on which the label was submitted

        Example use:
        def on_submit(self, label: Label, project_id: str, asset_id: str):
            json_response = label.get('jsonResponse')
            if label_is_respecting_business_rule(json_response):
                return
            else:
                self.kili.send_back_to_queue(asset_ids=[asset_id])

        """
        # pylint: disable=unused-argument
        self.logger.warning("Method not implemented. Define a custom on_submit on your plugin")
        pass  # pylint: disable=unnecessary-pass

    def on_review(
        self,
        label: Label,
        asset_id: str,
    ) -> None:
        """
        Handler for the review action, triggered when a default label is reviewed on Kili
        :param label: label submitted to Kili
        :param asset_id: id of the asset on which the label was submitted

        Example use:
        def on_review(self, label: Label, project_id: str, asset_id: str):
            json_response = label.get('jsonResponse')
            if label_is_respecting_business_rule(json_response):
                return
            else:
                self.kili.send_back_to_queue(asset_ids=[asset_id])
        """
        # pylint: disable=unused-argument
        self.logger.warning("Method not implemented. Define a custom on_review on your plugin")
        pass  # pylint: disable=unnecessary-pass



"""
Functions to import assets into an IMAGE project
"""
import mimetypes
import os
from typing import List

from .base import BaseAssetImporter, BatchParams, ContentBatchImporter
from .constants import LARGE_IMAGE_THRESHOLD_SIZE
from .types import AssetLike


class ImageDataImporter(BaseAssetImporter):
    """
    Class for importing assets into an IMAGE project
    """

    def import_assets(self, assets: List[AssetLike]):
        """
        Import IMAGE assets into Kili.
        """
        is_hosted = self.is_hosted_content(assets)
        if not is_hosted:
            assets = self.filter_local_assets(assets, self.raise_error)
        assets = self.filter_duplicate_external_ids(assets)
        sync_assets, async_assets = self.split_asset_by_upload_type(assets, is_hosted)
        results = []
        if len(sync_assets) > 0:
            sync_batch_params = BatchParams(is_hosted=is_hosted, is_asynchronous=False)
            batch_importer = ContentBatchImporter(
                self.auth, self.project_params, sync_batch_params, self.pbar
            )
            result = self.import_assets_by_batch(sync_assets, batch_importer)
            results.append(result)
        if len(async_assets) > 0:
            async_batch_params = BatchParams(is_hosted=is_hosted, is_asynchronous=True)
            batch_importer = ContentBatchImporter(
                self.auth, self.project_params, async_batch_params, self.pbar
            )
            result = self.import_assets_by_batch(async_assets, batch_importer)
            results.append(result)
        return results[0]

    @staticmethod
    def split_asset_by_upload_type(assets: List[AssetLike], is_hosted: bool):
        """
        Split assets into two groups, assets to to imported synchronously or asynchronously
        """
        if is_hosted:
            return assets, []
        sync_assets, async_assets = [], []
        for asset in assets:
            path = asset.get("content")
            assert path
            assert isinstance(path, str)
            mime_type, _ = mimetypes.guess_type(path.lower())
            is_large_image = os.path.getsize(path) >= LARGE_IMAGE_THRESHOLD_SIZE
            is_tiff = mime_type == "image/tiff"
            if is_large_image or is_tiff:
                async_assets.append(asset)
            else:
                sync_assets.append(asset)
        return sync_assets, async_assets



"""
Queries of label queries
"""


def gql_labels(fragment):
    """
    Return the GraphQL labels query
    """
    return f"""
query ($where: LabelWhere!, $first: PageSize!, $skip: Int!) {{
  data: labels(where: $where, first: $first, skip: $skip) {{
    {fragment}
  }}
}}
"""


GQL_LABELS_COUNT = """
query($where: LabelWhere!) {
  data: countLabels(where: $where)
}
"""



"""
Functions to get logs of a plugin
"""
from datetime import datetime
from typing import Dict, Optional

from kili.authentication import KiliAuth
from kili.graphql.operations.plugins.queries import GQL_GET_PLUGIN_LOGS
from kili.helpers import format_result


def get_logs(
    auth: KiliAuth,
    plugin: Dict[str, str],
    start_date: Optional[datetime] = None,
    limit: Optional[int] = None,
    skip: Optional[int] = None,
):
    """
    Get logs of a plugin in Kili
    """

    variables = {
        "projectId": plugin["project_id"],
        "pluginName": plugin["plugin_name"],
        "limit": limit,
        "skip": skip,
    }

    if start_date:
        variables["createdAt"] = start_date.isoformat(sep="T", timespec="milliseconds") + "Z"

    result = auth.client.execute(GQL_GET_PLUGIN_LOGS, variables)
    return format_result("data", result)



"""
Class to upload a plugin
"""

import time
from pathlib import Path
from typing import Optional

from kili.authentication import KiliAuth
from kili.constants import mime_extensions_for_py_scripts
from kili.graphql.operations.plugins.mutations import (
    GQL_CREATE_PLUGIN,
    GQL_CREATE_PLUGIN_RUNNER,
    GQL_GENERATE_UPDATE_URL,
    GQL_UPDATE_PLUGIN_RUNNER,
)
from kili.graphql.operations.plugins.queries import GQL_GET_PLUGIN_RUNNER_STATUS
from kili.helpers import format_result, get_data_type
from kili.services.plugins.tools import check_errors_plugin_upload
from kili.utils import bucket

from .helpers import get_logger

NUMBER_TRIES_RUNNER_STATUS = 20


def check_file_is_py(path: Path, verbose: bool = True) -> bool:
    """
    Returns true if the mime type of the file corresponds to a python file
    """
    mime_type = get_data_type(path.as_posix())

    if not (mime_extensions_for_py_scripts and mime_type):
        return False

    correct_mime_type = mime_type in mime_extensions_for_py_scripts

    if verbose and not correct_mime_type:
        print(
            f"File mime type for {path} is {mime_type} and does not correspond"
            "to the type of the project. "
            f"File mime type should be one of {mime_extensions_for_py_scripts}"
        )
    return correct_mime_type


class PluginUploader:  # pylint: disable=too-few-public-methods
    """
    Class to upload a plugin
    """

    def __init__(
        self, auth: KiliAuth, file_path: str, plugin_name: Optional[str], verbose: bool
    ) -> None:
        self.auth = auth
        self.file_path = Path(file_path)
        self.plugin_name = plugin_name or self.file_path.name
        self.verbose = verbose

    def _retrieve_script(self) -> str:
        """
        Retrieve script from file_path and execute it
        to prevent an upload with indentation errors
        """
        if not check_file_is_py(self.file_path, self.verbose):
            raise ValueError("Wrong file format. ")

        path = Path(self.file_path)

        with path.open("r", encoding="utf-8") as file:
            source_code = file.read()

        return source_code

    def _parse_script(self, source_code: str):
        """
        Method to detect indentation errors in the script
        """
        logger = get_logger()

        # We execute the source code to prevent the upload of a file with SyntaxError
        logger.info(f"Executing {self.file_path.name}...")
        # pylint: disable=exec-used
        exec(source_code, globals())
        logger.info(f"Done executing {self.file_path.name}!")

    @staticmethod
    def _upload_file(source_code: str, url: str):
        """
        Upload a file to a signed url and returns the url with the file_id
        """
        bucket.upload_data_via_rest(url, source_code.encode("utf-8"), "text/x-python")

    def _retrieve_upload_url(self, is_updating_plugin: bool) -> str:
        """
        Retrieve an upload url from the backend
        """
        variables = {"pluginName": self.plugin_name}

        if is_updating_plugin:
            result = self.auth.client.execute(GQL_GENERATE_UPDATE_URL, variables)
        else:
            result = self.auth.client.execute(GQL_CREATE_PLUGIN, variables)

        check_errors_plugin_upload(result, self.file_path, self.plugin_name)
        upload_url = format_result("data", result)
        return upload_url

    def _upload_script(self, is_updating_plugin: bool):
        """
        Upload a script to Kili bucket
        """

        upload_url = self._retrieve_upload_url(is_updating_plugin)

        source_code = self._retrieve_script()
        self._parse_script(source_code)
        self._upload_file(source_code, upload_url)

    def _create_plugin_runner(self):
        """
        Create plugin's runner
        """

        variables = {"pluginName": self.plugin_name}

        result = self.auth.client.execute(GQL_CREATE_PLUGIN_RUNNER, variables)
        return format_result("data", result)

    def _check_plugin_runner_status(self, update=False):
        """
        Check the status of a plugin's runner until it is active
        """

        logger = get_logger()

        action = "updated" if update else "created"

        logger.info(f"Plugin is being {action}... This should take approximately 3 minutes.")

        n_tries = 0
        status = None

        while n_tries < NUMBER_TRIES_RUNNER_STATUS:
            time.sleep(15)

            status = self.get_plugin_runner_status()

            if status == "ACTIVE":
                break

            logger.info(f"Status : {status}...")

            n_tries += 1

        if status == "DEPLOYING" and n_tries == 20:
            raise Exception(
                f"""We could not check your plugin was deployed in time.
Please check again the status of the plugin after some minutes with the command : \
kili.get_plugin_status("{self.plugin_name}").
If the status is different than DEPLOYING or ACTIVE, please check your plugin's code and try to \
overwrite the plugin with a new version of the code (you can use kili.update_plugin() for that)."""
            )

        if status != "ACTIVE":
            raise Exception(
                """There was some error during the creation of the plugin. \
Please check your plugin's code and try to overwrite the plugin with a new version of the \
code (you can use kili.update_plugin() for that)."""
            )

        message = f"Plugin {action} successfully"
        logger.info(message)

        return message

    def get_plugin_runner_status(self):
        """
        Get the status of a plugin's runner
        """

        variables = {"name": self.plugin_name}

        result = self.auth.client.execute(GQL_GET_PLUGIN_RUNNER_STATUS, variables)

        return format_result("data", result)

    def create_plugin(self):
        """
        Create a plugin in Kili
        """

        self._upload_script(False)

        self._create_plugin_runner()

        return self._check_plugin_runner_status()

    def _update_plugin_runner(self):
        """
        Update plugin's runner
        """
        variables = {"pluginName": self.plugin_name}

        result = self.auth.client.execute(GQL_UPDATE_PLUGIN_RUNNER, variables)
        return format_result("data", result)

    def update_plugin(self):
        """
        Update a plugin in Kili
        """

        self._upload_script(True)

        self._update_plugin_runner()

        return self._check_plugin_runner_status(update=True)



"""
Common code for the yolo exporter.
"""

import csv
import json
import logging
from pathlib import Path
from typing import Dict, List, Set, Tuple

from kili.services.export.exceptions import NoCompatibleJobError, NotCompatibleOptions
from kili.services.export.format.base import AbstractExporter
from kili.services.export.repository import AbstractContentRepository, DownloadError
from kili.services.export.types import JobCategory, LabelFormat, YoloAnnotation
from kili.utils.tempfile import TemporaryDirectory
from kili.utils.tqdm import tqdm


class YoloExporter(AbstractExporter):
    """
    Common code for Yolo exporters.
    """

    def _check_arguments_compatibility(self):
        if self.single_file:
            raise NotCompatibleOptions(
                f"The label format {self.label_format} can not be exported into a single file."
            )

    def process_and_save(self, assets: List[Dict], output_filename: Path) -> None:
        """
        Yolo specific process and save
        """
        if self.split_option == "merged":
            return self._process_and_save_merge(assets, output_filename)
        return self._process_and_save_split(assets, output_filename)

    def _process_and_save_split(self, assets: List[Dict], output_filename: Path) -> None:
        self.logger.info("Exporting to yolo format split...")

        json_interface, ml_task, tool = self.get_project_and_init()
        categories_by_job = self._get_categories_by_job(json_interface, ml_task, tool)
        if not categories_by_job:
            raise NoCompatibleJobError(
                f"Error: There is no job in project {self.project_id} "
                f"that can be converted to the {self.label_format} format."
            )

        with TemporaryDirectory() as root_folder:
            images_folder = root_folder / self.project_id / "images"
            images_folder.mkdir(parents=True)
            self._write_jobs_labels_into_split_folders(
                assets,
                categories_by_job,
                root_folder,
                images_folder,
            )
            self.create_readme_kili_file(root_folder)
            self.make_archive(root_folder, output_filename)

        self.logger.warning(output_filename)

    def _process_and_save_merge(self, assets: List[Dict], output_filename: Path) -> None:
        self.logger.info("Exporting to yolo format merged...")
        json_interface, ml_task, tool = self.get_project_and_init()
        merged_categories_id = self._get_merged_categories(json_interface, ml_task, tool)
        if not merged_categories_id:
            raise NoCompatibleJobError(
                f"Error: There is no job in project {self.project_id} "
                f"that can be converted to the {self.label_format} format."
            )

        with TemporaryDirectory() as root_folder:
            base_folder = root_folder / self.project_id
            images_folder = base_folder / "images"
            labels_folder = base_folder / "labels"
            images_folder.mkdir(parents=True)
            labels_folder.mkdir(parents=True)
            self._write_labels_into_single_folder(
                assets,
                merged_categories_id,
                labels_folder,
                images_folder,
                base_folder,
            )
            self.create_readme_kili_file(root_folder)
            self.make_archive(root_folder, output_filename)

        self.logger.warning(output_filename)

    @classmethod
    def _get_categories_by_job(
        cls, json_interface: Dict, ml_task: str, tool: str
    ) -> Dict[str, Dict[str, JobCategory]]:
        """
        Return a dictionary of JobCategory instances by category full name and job id.
        """
        categories_by_job: Dict[str, Dict[str, JobCategory]] = {}
        for job_id, job in json_interface.get("jobs", {}).items():
            if (
                job.get("mlTask") != ml_task
                or tool not in job.get("tools", [])
                or job.get("isModel")
            ):
                continue
            categories: Dict[str, JobCategory] = {}
            for cat_id, category in enumerate(job.get("content", {}).get("categories", {})):
                categories[get_category_full_name(job_id, category)] = JobCategory(
                    category_name=category, id=cat_id, job_id=job_id
                )
            categories_by_job[job_id] = categories
        return categories_by_job

    def _write_labels_into_single_folder(
        self,
        assets: List[Dict],
        categories_id: Dict[str, JobCategory],
        labels_folder: Path,
        images_folder: Path,
        base_folder: Path,
    ):  # pylint: disable=too-many-arguments
        """
        Write all the labels into a single folder.
        """
        _write_class_file(base_folder, categories_id, self.label_format)

        remote_content = []
        video_metadata = {}

        for asset in tqdm(assets, disable=self.disable_tqdm):
            asset_remote_content, video_filenames = _process_asset(
                asset, images_folder, labels_folder, categories_id, self.content_repository
            )
            if video_filenames:
                video_metadata[asset["externalId"]] = video_filenames
            remote_content.extend(asset_remote_content)

        if video_metadata:
            _write_video_metadata_file(video_metadata, base_folder)

        if len(remote_content) > 0:
            _write_remote_content_file(remote_content, images_folder)

    def _write_jobs_labels_into_split_folders(
        self,
        assets: List[Dict],
        categories_by_job: Dict[str, Dict[str, JobCategory]],
        root_folder: Path,
        images_folder: Path,
    ) -> None:
        """
        Write assets into split folders.
        """
        for job_id, category_ids in categories_by_job.items():
            base_folder = root_folder / self.project_id / job_id
            labels_folder = base_folder / "labels"
            labels_folder.mkdir(parents=True)

            self._write_labels_into_single_folder(
                assets,
                category_ids,
                labels_folder,
                images_folder,
                base_folder,
            )

    @classmethod
    def _get_merged_categories(
        cls, json_interface: Dict, ml_task: str, tool: str
    ) -> Dict[str, JobCategory]:
        """
        Return a dictionary of JobCategory instances by category full name.
        """
        cat_number = 0
        merged_categories_id: Dict[str, JobCategory] = {}
        for job_id, job in json_interface.get("jobs", {}).items():
            if (
                job.get("mlTask") != ml_task
                or tool not in job.get("tools", [])
                or job.get("isModel")
            ):
                continue
            for category in job.get("content", {}).get("categories", {}):
                merged_categories_id[get_category_full_name(job_id, category)] = JobCategory(
                    category_name=category, id=cat_number, job_id=job_id
                )
                cat_number += 1

        return merged_categories_id


class _LabelFrames:
    """
    Holds asset frames data.
    """

    @staticmethod
    def from_asset(asset, job_ids) -> "_LabelFrames":
        """
        Instantiate the label frames from the asset. It handles the case when there are several
        frames by label or a single one.
        """
        frames = {}
        number_of_frames = 0
        is_frame_group = False
        if "jsonResponse" in asset["latestLabel"]:
            number_of_frames = len(asset["latestLabel"]["jsonResponse"])
            for idx in range(number_of_frames):
                if str(idx) in asset["latestLabel"]["jsonResponse"]:
                    is_frame_group = True
                    frame_asset = asset["latestLabel"]["jsonResponse"][str(idx)]
                    for job_id in job_ids:
                        if (
                            job_id in frame_asset
                            and "annotations" in frame_asset[job_id]
                            and frame_asset[job_id]["annotations"]
                        ):
                            frames[idx] = {"latestLabel": {"jsonResponse": frame_asset}}
                            break

        if not frames:
            frames[-1] = asset
        return _LabelFrames(frames, number_of_frames, is_frame_group, asset["externalId"])

    def __init__(
        self, frames: Dict[int, Dict], number_frames: int, is_frame_group: bool, external_id: str
    ) -> None:
        self.frames: Dict[int, Dict] = frames
        self.number_frames: int = number_frames
        self.is_frame_group: bool = is_frame_group
        self.external_id: str = external_id

    def get_leading_zeros(self) -> int:
        """
        Get leading zeros for file name building
        """
        return len(str(self.number_frames))

    def get_label_filename(self, idx: int) -> str:
        """
        Get label filemame for index
        """
        return f"{self.external_id}_{str(idx + 1).zfill(self.get_leading_zeros())}"


def _convert_from_kili_to_yolo_format(
    job_id: str, label: Dict, category_ids: Dict[str, JobCategory]
) -> List[YoloAnnotation]:
    # pylint: disable=too-many-locals
    """
    Extract formatted annotations from labels and save the zip in the buckets.
    """
    if label is None or "jsonResponse" not in label:
        return []
    json_response = label["jsonResponse"]
    if not (job_id in json_response and "annotations" in json_response[job_id]):
        return []
    annotations = json_response[job_id]["annotations"]
    converted_annotations = []
    for annotation in annotations:
        category_idx: JobCategory = category_ids[
            get_category_full_name(job_id, annotation["categories"][0]["name"])
        ]
        if "boundingPoly" not in annotation:
            continue
        bounding_poly = annotation["boundingPoly"]
        if len(bounding_poly) < 1 or "normalizedVertices" not in bounding_poly[0]:
            continue
        normalized_vertices = bounding_poly[0]["normalizedVertices"]
        x_s = [vertice["x"] for vertice in normalized_vertices]
        y_s = [vertice["y"] for vertice in normalized_vertices]
        x_min, y_min = min(x_s), min(y_s)
        x_max, y_max = max(x_s), max(y_s)
        _x_, _y_ = (x_max + x_min) / 2, (y_max + y_min) / 2
        _w_, _h_ = x_max - x_min, y_max - y_min

        converted_annotations.append((category_idx.id, _x_, _y_, _w_, _h_))
    return converted_annotations


def get_category_full_name(job_id: str, category_name: str):
    """
    Return a full name to identify uniquely a category
    """
    return f"{job_id}__{category_name}"


def _process_asset(
    asset: Dict,
    images_folder: Path,
    labels_folder: Path,
    category_ids: Dict[str, JobCategory],
    content_repository: AbstractContentRepository,
) -> Tuple[List[Tuple[str, str, str]], List[str]]:
    # pylint: disable=too-many-locals, too-many-branches
    """
    Process an asset for all job_ids of category_ids.
    """
    asset_remote_content = []
    job_ids = set(map(lambda job_category: job_category.job_id, category_ids.values()))

    label_frames = _LabelFrames.from_asset(asset, job_ids)

    content_frames = content_repository.get_content_frames_paths(asset)

    video_filenames = []

    for idx, frame in label_frames.frames.items():
        if label_frames.is_frame_group:
            filename = label_frames.get_label_filename(idx)
            video_filenames.append(filename)
        else:
            filename = asset["externalId"]

        frame_labels = _get_frame_labels(frame, job_ids, category_ids)

        if not frame_labels:
            continue

        _write_labels_to_file(labels_folder, filename, frame_labels)

        content_frame = content_frames[idx] if content_frames else asset["content"]
        if content_repository.is_serving(content_frame):
            if content_frames and not label_frames.is_frame_group:
                try:
                    _write_content_frame_to_file(
                        content_frame, images_folder, filename, content_repository
                    )
                except DownloadError as download_error:
                    asset_id = asset["id"]
                    logging.warning("for asset %s: %s", asset_id, str(download_error))
        else:
            asset_remote_content.append([asset["externalId"], content_frame, f"{filename}.txt"])

    if (
        not content_frames
        and label_frames.is_frame_group
        and content_repository.is_serving(asset["content"])
    ):
        raise NotImplementedError("Export of annotations on videos is not supported yet.")

    return asset_remote_content, video_filenames


def _write_class_file(
    folder: Path, category_ids: Dict[str, JobCategory], label_format: LabelFormat
):
    """
    Create a file that contains meta information about the export, depending of Yolo version
    """
    if label_format == "yolo_v4":
        with (folder / "classes.txt").open("wb") as fout:
            for job_category in category_ids.values():
                fout.write(f"{job_category.id} {job_category.category_name}\n".encode())
    if label_format == "yolo_v5":
        with (folder / "data.yaml").open("wb") as fout:
            fout.write("names:\n".encode())
            categories = ""
            for ind, job_category in enumerate(category_ids.values()):
                fout.write(f"  {ind}: {job_category.category_name}\n".encode())
    if label_format == "yolo_v7":
        with (folder / "data.yaml").open("wb") as fout:
            categories = ""
            for job_category in category_ids.values():
                categories += f"'{job_category.category_name}', "
            fout.write(f"nc: {len(category_ids.items())}\n".encode())
            fout.write(f"names: [{categories[:-2]}]\n".encode())


def _get_frame_labels(
    frame: Dict, job_ids: Set[str], category_ids: Dict[str, JobCategory]
) -> List[YoloAnnotation]:
    annotations = []
    for job_id in job_ids:
        job_annotations = _convert_from_kili_to_yolo_format(
            job_id, frame["latestLabel"], category_ids
        )
        annotations += job_annotations

    return annotations


def _write_content_frame_to_file(
    url_content_frame: str,
    images_folder: Path,
    filename: str,
    content_repository: AbstractContentRepository,
):
    content_iterator = content_repository.get_content_stream(url_content_frame, 1024)
    with (images_folder / f"{filename}.jpg").open("wb") as fout:
        for block in content_iterator:
            if not block:
                break
            fout.write(block)


def _write_labels_to_file(
    labels_folder: Path, filename: str, annotations: List[YoloAnnotation]
) -> None:
    with (labels_folder / f"{filename}.txt").open("wb") as fout:
        for category_idx, _x_, _y_, _w_, _h_ in annotations:
            fout.write(f"{category_idx} {_x_} {_y_} {_w_} {_h_}\n".encode())


def _write_video_metadata_file(video_metadata: Dict, base_folder: Path) -> None:
    """
    Write video metadata file
    """
    video_metadata_json = json.dumps(video_metadata, sort_keys=True, indent=4)
    if video_metadata_json is not None:
        with (base_folder / "video_meta.json").open("wb") as output_file:
            output_file.write(video_metadata_json.encode("utf-8"))


def _write_remote_content_file(remote_content: List[str], images_folder: Path) -> None:
    """
    Write remote content file
    """
    remote_content_header = ["external id", "url", "label file"]
    with (images_folder / "remote_assets.csv").open("w", encoding="utf8") as file:
        writer = csv.writer(file)
        writer.writerow(remote_content_header)
        writer.writerows(remote_content)



"""
Api key mutations
"""

from typing import Dict

from typeguard import typechecked

from ...helpers import format_result
from .queries import GQL_APPEND_TO_API_KEYS


class MutationsApiKey:  # pylint: disable=too-few-public-methods
    """Set of User mutations."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @typechecked
    def append_to_api_keys(self, api_key: str, name: str) -> Dict[str, str]:
        """Create an api key to connect to the API.

        Args:
            api_key: A new api key to connect with.
            name: A name used to describe the api key.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {
            "data": {"key": api_key, "name": name},
            "where": {"email": self.auth.user_email},
        }
        result = self.auth.client.execute(GQL_APPEND_TO_API_KEYS, variables)
        return format_result("data", result)



"""
Export-related exceptions
"""


class DownloadError(Exception):
    """
    Exception thrown when the contents cannot be downloaded.
    """


class NoCompatibleJobError(Exception):
    """
    Exception thrown when there is no job compatible with the format.
    """


class NotCompatibleOptions(ValueError):
    """
    Exception thrown when two export options are not compatible
    """



"""
Asset mutations
"""

from typing import Any, Dict, List, Optional, Union

from typeguard import typechecked

from kili.helpers import format_result
from kili.mutations.asset.helpers import process_update_properties_in_assets_parameters
from kili.mutations.asset.queries import (
    GQL_ADD_ALL_LABELED_ASSETS_TO_REVIEW,
    GQL_DELETE_MANY_FROM_DATASET,
    GQL_SEND_BACK_ASSETS_TO_QUEUE,
    GQL_UPDATE_PROPERTIES_IN_ASSETS,
)
from kili.orm import Asset
from kili.services.asset_import import import_assets
from kili.utils.pagination import _mutate_from_paginated_call


class MutationsAsset:
    """
    Set of Asset mutations
    """

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @typechecked
    def append_many_to_dataset(
        self,
        project_id: str,
        content_array: Optional[List[str]] = None,
        external_id_array: Optional[List[str]] = None,
        id_array: Optional[List[str]] = None,
        is_honeypot_array: Optional[List[bool]] = None,
        status_array: Optional[List[str]] = None,
        json_content_array: Optional[List[List[Union[dict, str]]]] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: bool = False,
    ) -> Dict[str, str]:
        # pylint: disable=line-too-long
        """Append assets to a project.

        Args:
            project_id: Identifier of the project
            content_array: List of elements added to the assets of the project
                Must not be None except if you provide json_content_array.

                - For a `TEXT` project, the content can be either raw text, or URLs to TEXT assets.
                - For an `IMAGE` / `PDF` project, the content can be either URLs or paths to existing
                    images/pdf on your computer.
                - For a VIDEO project, the content can be either URLs pointing to videos hosted on a web server or paths to
                existing video files on your computer. If you want to import video from frames, look at the json_content
                section below.
                - For an `VIDEO_LEGACY` project, the content can be only be URLs
            external_id_array: List of external ids given to identify the assets.
                If None, random identifiers are created.
            is_honeypot_array:  Whether to use the asset for honeypot
            status_array: By default, all imported assets are set to `TODO`. Other options:
                `ONGOING`, `LABELED`, `REVIEWED`.
            json_content_array: Useful for `VIDEO` or `TEXT` projects only.

                - For `VIDEO` projects, each element is a sequence of frames, i.e. a
                    list of URLs to images or a list of paths to images.
                - For `TEXT` projects, each element is a json_content dict,
                    formatted according to documentation [on how to import
                rich-text assets](https://github.com/kili-technology/kili-python-sdk/blob/master/recipes/import_text_assets.ipynb)
            json_metadata_array: The metadata given to each asset should be stored in a json like dict with keys.

                - Add metadata visible on the asset with the following keys: `imageUrl`, `text`, `url`.
                    Example for one asset: `json_metadata_array = [{'imageUrl': '','text': '','url': ''}]`.
                - For VIDEO projects (and not VIDEO_LEGACY), you can specify a value with key 'processingParameters' to specify the sampling rate (default: 30).
                    Example for one asset: `json_metadata_array = [{'processingParameters': {'framesPlayedPerSecond': 10}}]`.
            disable_tqdm: If `True`, the progress bar will be disabled

        Returns:
            A result object which indicates if the mutation was successful, or an error message.

        Examples:
            >>> kili.append_many_to_dataset(
                    project_id=project_id,
                    content_array=['https://upload.wikimedia.org/wikipedia/en/7/7d/Lenna_%28test_image%29.png'])

        !!! example "Recipe"
            - For more detailed examples on how to import assets,
                see [the recipe](https://github.com/kili-technology/kili-python-sdk/blob/master/recipes/import_assets.ipynb)
                or [other examples](https://docs.kili-technology.com/recipes/importing-data) in our documentation.
            - For more detailed examples on how to import text assets,
                see [the recipe](https://github.com/kili-technology/kili-python-sdk/blob/master/recipes/import_text_assets.ipynb).
        """

        if content_array is None and json_content_array is None:
            raise ValueError("Variables content_array and json_content_array cannot be both None.")
        nb_data = (
            len(content_array)
            if content_array is not None
            else len(json_content_array)  # type:ignore
        )
        field_mapping = {
            "content": content_array,
            "json_content": json_content_array,
            "external_id": external_id_array,
            "id": id_array,
            "status": status_array,
            "json_metadata": json_metadata_array,
            "is_honeypot": is_honeypot_array,
        }
        assets = [{}] * nb_data
        for key, value in field_mapping.items():
            if value is not None:
                assets = [{**assets[i], key: value[i]} for i in range(nb_data)]
        result = import_assets(
            self.auth, project_id=project_id, assets=assets, disable_tqdm=disable_tqdm
        )
        return result

    @typechecked
    # pylint: disable=unused-argument
    def update_properties_in_assets(
        self,
        asset_ids: List[str],
        external_ids: Optional[List[str]] = None,
        priorities: Optional[List[int]] = None,
        json_metadatas: Optional[List[Union[dict, str]]] = None,
        consensus_marks: Optional[List[float]] = None,
        honeypot_marks: Optional[List[float]] = None,
        to_be_labeled_by_array: Optional[List[List[str]]] = None,
        contents: Optional[List[str]] = None,
        json_contents: Optional[List[str]] = None,
        status_array: Optional[List[str]] = None,
        is_used_for_consensus_array: Optional[List[bool]] = None,
        is_honeypot_array: Optional[List[bool]] = None,
    ) -> List[dict]:
        """Update the properties of one or more assets.

        Args:
            asset_ids : The asset IDs to modify
            external_ids: Change the external id of the assets
            priorities : You can change the priority of the assets
                By default, all assets have a priority of 0.
            json_metadatas: The metadata given to an asset should be stored
                in a json like dict with keys `imageUrl`, `text`, `url`:
                `json_metadata = {'imageUrl': '','text': '','url': ''}`
            consensus_marks: Should be between 0 and 1
            honeypot_marks: Should be between 0 and 1
            to_be_labeled_by_array: If given, each element of the list should contain the emails of
                the labelers authorized to label the asset.
            contents: - For a NLP project, the content can be directly in text format
                - For an Image / Video / Pdf project, the content must be hosted on a web server,
                and you point Kili to your data by giving the URLs
            json_contents: - For a NLP project, the `json_content`
                is a text formatted using RichText
                - For a Video project, the`json_content` is a json containg urls pointing
                    to each frame of the video.
            status_array: Each element should be in `TODO`, `ONGOING`, `LABELED`, `REVIEWED`
            is_used_for_consensus_array: Whether to use the asset to compute consensus kpis or not
            is_honeypot_array: Whether to use the asset for honeypot

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.update_properties_in_assets(
                    asset_ids=["ckg22d81r0jrg0885unmuswj8",
                        "ckg22d81s0jrh0885pdxfd03n"],
                    consensus_marks=[1, 0.7],
                    contents=[None, 'https://to/second/asset.png'],
                    external_ids=['external-id-of-your-choice-1',
                        'external-id-of-your-choice-2'],
                    honeypot_marks=[0.8, 0.5],
                    is_honeypot_array=[True, True],
                    is_used_for_consensus_array=[True, False],
                    priorities=[None, 2],
                    status_array=['LABELED', 'REVIEWED'],
                    to_be_labeled_by_array=[
                        ['test+pierre@kili-technology.com'], None],
            )
        """

        saved_args = locals()
        parameters = {
            k: v
            for (k, v) in saved_args.items()
            if k
            in [
                "asset_ids",
                "external_ids",
                "priorities",
                "json_metadatas",
                "consensus_marks",
                "honeypot_marks",
                "to_be_labeled_by_array",
                "contents",
                "json_contents",
                "status_array",
                "is_used_for_consensus_array",
                "is_honeypot_array",
            ]
        }
        properties_to_batch = process_update_properties_in_assets_parameters(parameters)

        def generate_variables(batch: Dict) -> Dict:
            data = {
                "externalId": batch["external_ids"],
                "priority": batch["priorities"],
                "jsonMetadata": batch["json_metadatas"],
                "consensusMark": batch["consensus_marks"],
                "honeypotMark": batch["honeypot_marks"],
                "toBeLabeledBy": batch["to_be_labeled_by_array"],
                "shouldResetToBeLabeledBy": batch["should_reset_to_be_labeled_by_array"],
                "content": batch["contents"],
                "jsonContent": batch["json_contents"],
                "status": batch["status_array"],
                "isUsedForConsensus": batch["is_used_for_consensus_array"],
                "isHoneypot": batch["is_honeypot_array"],
            }
            data_array = [dict(zip(data, t)) for t in zip(*data.values())]
            return {
                "whereArray": [{"id": asset_id} for asset_id in batch["asset_ids"]],
                "dataArray": data_array,
            }

        results = _mutate_from_paginated_call(
            self,
            properties_to_batch,
            generate_variables,
            GQL_UPDATE_PROPERTIES_IN_ASSETS,
        )
        formated_results = [format_result("data", result, Asset) for result in results]
        return [item for batch_list in formated_results for item in batch_list]

    @typechecked
    def delete_many_from_dataset(self, asset_ids: List[str]) -> Asset:
        """Delete assets from a project.

        Args:
            asset_ids: The list of identifiers of the assets to delete.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        properties_to_batch: Dict[str, Optional[List[Any]]] = {"asset_ids": asset_ids}

        def generate_variables(batch):
            return {"where": {"idIn": batch["asset_ids"]}}

        results = _mutate_from_paginated_call(
            self, properties_to_batch, generate_variables, GQL_DELETE_MANY_FROM_DATASET
        )
        return format_result("data", results[0], Asset)

    @typechecked
    def add_to_review(self, asset_ids: List[str]) -> Dict:
        """Add assets to review.

        !!! warning
            Assets without any label will be ignored.

        Args:
            asset_ids: The asset IDs to add to review

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.add_to_review(
                    asset_ids=[
                        "ckg22d81r0jrg0885unmuswj8",
                        "ckg22d81s0jrh0885pdxfd03n"
                        ],
        """
        properties_to_batch: Dict[str, Optional[List[Any]]] = {"asset_ids": asset_ids}

        def generate_variables(batch):
            return {"where": {"idIn": batch["asset_ids"]}}

        results = _mutate_from_paginated_call(
            self,
            properties_to_batch,
            generate_variables,
            GQL_ADD_ALL_LABELED_ASSETS_TO_REVIEW,
        )
        return format_result("data", results[0])

    @typechecked
    def send_back_to_queue(self, asset_ids: List[str]) -> Dict[str, str]:
        """Send assets back to queue.

        Args:
            asset_ids: The asset IDs to add to review

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.send_back_to_queue(
                    asset_ids=[
                        "ckg22d81r0jrg0885unmuswj8",
                        "ckg22d81s0jrh0885pdxfd03n"
                        ],
        """
        properties_to_batch: Dict[str, Optional[List[Any]]] = {"asset_ids": asset_ids}

        def generate_variables(batch):
            return {"where": {"idIn": batch["asset_ids"]}}

        results = _mutate_from_paginated_call(
            self, properties_to_batch, generate_variables, GQL_SEND_BACK_ASSETS_TO_QUEUE
        )
        return format_result("data", results[0])



"""Label queries."""

from typing import Dict, Generator, Iterable, List, Optional, Union, cast

import pandas as pd
from typeguard import typechecked

from kili import services
from kili.constants import NO_ACCESS_RIGHT
from kili.helpers import format_result, fragment_builder, validate_category_search_query
from kili.queries.asset import QueriesAsset
from kili.queries.label.queries import GQL_LABELS_COUNT, gql_labels
from kili.queries.project import QueriesProject
from kili.services.export.exceptions import NoCompatibleJobError
from kili.services.export.types import LabelFormat, SplitOption
from kili.services.types import ProjectId
from kili.types import Label as LabelType
from kili.utils.pagination import row_generator_from_paginated_calls


class QueriesLabel:
    """Set of Label queries."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @typechecked
    def labels(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[List[str]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: List[str] = [
            "author.email",
            "author.id",
            "id",
            "jsonResponse",
            "labelType",
            "secondsToLabel",
        ],
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        type_in: Optional[List[str]] = None,
        user_id: Optional[str] = None,
        disable_tqdm: bool = False,
        as_generator: bool = False,
        category_search: Optional[str] = None,
    ) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Get a label list or a label generator from a project based on a set of criteria.

        Args:
            project_id: Identifier of the project.
            asset_id: Identifier of the asset.
            asset_status_in: Returned labels should have a status that belongs to that list, if given.
                Possible choices : `TODO`, `ONGOING`, `LABELED` or `REVIEWED`
            asset_external_id_in: Returned labels should have an external id that belongs to that list, if given.
            author_in: Returned labels should have a label whose status belongs to that list, if given.
            created_at: Returned labels should have a label whose creation date is equal to this date.
            created_at_gte: Returned labels should have a label whose creation date is greater than this date.
            created_at_lte: Returned labels should have a label whose creation date is lower than this date.
            fields: All the fields to request among the possible fields for the labels.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#label) for all possible fields.
            first: Maximum number of labels to return.
            honeypot_mark_gte: Returned labels should have a label whose honeypot is greater than this number.
            honeypot_mark_lte: Returned labels should have a label whose honeypot is lower than this number.
            id_contains: Filters out labels not belonging to that list. If empty, no filtering is applied.
            label_id: Identifier of the label.
            skip: Number of labels to skip (they are ordered by their date of creation, first to last).
            type_in: Returned labels should have a label whose type belongs to that list, if given.
            user_id: Identifier of the user.
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the labels is returned.

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            A result object which contains the query if it was successful, else an error message.

        Examples:
            >>> kili.labels(project_id=project_id, fields=['jsonResponse', 'labelOf.externalId']) # returns a list of all labels of a project and their assets external ID
            >>> kili.labels(project_id=project_id, fields=['jsonResponse'], as_generator=True) # returns a generator of all labels of a project

        !!! example "How to filter based on label categories"
            The search query is composed of logical expressions following this format:

                [job_name].[category_name].count [comparaison_operator] [value]
            where:

            - `[job_name]` is the name of the job in the interface
            - `[category_name]` is the name of the category in the interface for this job
            - `[comparaison_operator]` can be one of: [`==`, `>=`, `<=`, `<`, `>`]
            - `[value]` is an integer that represents the count of such objects of the given category in the label

            These operations can be separated by OR and AND operators

            Example:

                category_search = `JOB_CLASSIF.CATEGORY_A.count > 0`
                category_search = `JOB_CLASSIF.CATEGORY_A.count > 0 OR JOB_NER.CATEGORY_B.count > 0`
                category_search = `(JOB_CLASSIF.CATEGORY_A.count > 0 OR JOB_NER.CATEGORY_B.count > 0) AND JOB_BBOX.CATEGORY_C.count > 10`
        """

        saved_args = locals()
        count_args = {
            k: v
            for (k, v) in saved_args.items()
            if k
            not in [
                "as_generator",
                "disable_tqdm",
                "fields",
                "first",
                "id_contains",
                "self",
                "skip",
                "message",
            ]
        }

        # using tqdm with a generator is messy, so it is always disabled
        disable_tqdm = disable_tqdm or as_generator

        if category_search:
            validate_category_search_query(category_search)

        payload_query = {
            "where": {
                "id": label_id,
                "asset": {
                    "id": asset_id,
                    "externalIdStrictlyIn": asset_external_id_in,
                    "statusIn": asset_status_in,
                },
                "project": {
                    "id": project_id,
                },
                "user": {
                    "id": user_id,
                },
                "createdAt": created_at,
                "createdAtGte": created_at_gte,
                "createdAtLte": created_at_lte,
                "authorIn": author_in,
                "honeypotMarkGte": honeypot_mark_gte,
                "honeypotMarkLte": honeypot_mark_lte,
                "idIn": id_contains,
                "search": category_search,
                "typeIn": type_in,
            },
        }

        labels_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_labels,
            count_args,
            self._query_labels,
            payload_query,
            fields,
            disable_tqdm,
        )

        if as_generator:
            return labels_generator
        return list(labels_generator)

    def _query_labels(
        self, skip: int, first: int, payload: dict, fields: List[str]
    ) -> Iterable[Dict]:
        payload.update({"skip": skip, "first": first})
        _gql_labels = gql_labels(fragment_builder(fields, LabelType))
        result = self.auth.client.execute(_gql_labels, payload)
        return format_result("data", result, _object=List[LabelType])  # type:ignore

    # pylint: disable=dangerous-default-value
    @typechecked
    def export_labels_as_df(
        self,
        project_id: str,
        fields: List[str] = [
            "author.email",
            "author.id",
            "createdAt",
            "id",
            "labelType",
        ],
        asset_fields: List[str] = ["externalId"],
    ) -> pd.DataFrame:
        # pylint: disable=line-too-long
        """Get the labels of a project as a pandas DataFrame.

        Args:
            project_id: Identifier of the project
            fields: All the fields to request among the possible fields for the labels.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#label) for all possible fields.
            asset_fields: All the fields to request among the possible fields for the assets.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#asset) for all possible fields.

        Returns:
            pandas DataFrame containing the labels.
        """

        projects = QueriesProject(self.auth).projects(project_id)
        assert len(list(projects)) == 1, NO_ACCESS_RIGHT
        assets = QueriesAsset(self.auth).assets(
            project_id=project_id,
            fields=asset_fields + ["labels." + field for field in fields],
        )
        labels = [
            dict(
                label,
                **dict((f"asset_{key}", asset[key]) for key in asset if key != "labels"),
            )
            for asset in assets
            for label in asset["labels"]
        ]
        labels_df = pd.DataFrame(labels)
        return labels_df

    @typechecked
    def count_labels(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[List[str]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        label_id: Optional[str] = None,
        type_in: Optional[List[str]] = None,
        user_id: Optional[str] = None,
        category_search: Optional[str] = None,
    ) -> int:
        # pylint: disable=line-too-long
        """Get the number of labels for the given parameters.

        Args:
            asset_id: Identifier of the asset.
            asset_status_in: Returned labels should have a status that belongs to that list, if given.
                Possible choices : `TODO`, `ONGOING`, `LABELED` or `REVIEWED`
            asset_external_id_in: Returned labels should have an external id that belongs to that list, if given.
            author_in: Returned labels should have a label whose status belongs to that list, if given.
            created_at: Returned labels should have a label whose creation date is equal to this date.
            created_at_gte: Returned labels should have a label whose creation date is greater than this date.
            created_at_lte: Returned labels should have a label whose creation date is lower than this date.
            honeypot_mark_gte: Returned labels should have a label whose honeypot is greater than this number.
            honeypot_mark_lte: Returned labels should have a label whose honeypot is lower than this number.
            label_id: Identifier of the label.
            project_id: Identifier of the project.
            type_in: Returned labels should have a label whose type belongs to that list, if given.
            user_id: Identifier of the user.

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            The number of labels with the parameters provided
        """

        if category_search:
            validate_category_search_query(category_search)

        variables = {
            "where": {
                "id": label_id,
                "asset": {
                    "id": asset_id,
                    "externalIdStrictlyIn": asset_external_id_in,
                    "statusIn": asset_status_in,
                },
                "project": {
                    "id": project_id,
                },
                "user": {
                    "id": user_id,
                },
                "createdAt": created_at,
                "createdAtGte": created_at_gte,
                "createdAtLte": created_at_lte,
                "authorIn": author_in,
                "honeypotMarkGte": honeypot_mark_gte,
                "honeypotMarkLte": honeypot_mark_lte,
                "search": category_search,
                "typeIn": type_in,
            }
        }
        result = self.auth.client.execute(GQL_LABELS_COUNT, variables)
        return format_result("data", result, int)

    def export_labels(
        self,
        project_id: str,
        filename: str,
        fmt: LabelFormat,
        asset_ids: Optional[List[str]] = None,
        layout: SplitOption = "split",
        single_file: bool = False,
        disable_tqdm: bool = False,
    ):
        """
        Export the project labels with the requested format into the requested output path.

        Args:
            filename: Relative or full path of the archive that will contain
                the exported data.
            fmt: Format of the exported labels.
            asset_ids: Optional list of the assets from which to export the labels.
            layout: Layout of the exported files: "split" means there is one folder
                per job, "merged" that there is one folder with every labels.
            disable_tqdm: Disable the progress bar if True.

        !!! Info
            The supported formats are:

            - Yolo V4, V5, V7 for object detection tasks (bounding box)

            - Kili for all tasks.

            - COCO for semantic segmentation tasks (bounding box and semantic segmentation)

            - Pascal VOC (coming soon) for object detection tasks.

        !!! Example
            ```
            from kili.client import Kili
            kili = Kili()
            kili.export_labels("your_project_id", "export.zip", "yolo_v4")
            ```
        """

        try:
            services.export_labels(
                self,
                asset_ids=asset_ids,
                project_id=cast(ProjectId, project_id),
                export_type="latest",
                label_format=fmt,
                split_option=layout,
                single_file=single_file,
                output_file=filename,
                disable_tqdm=disable_tqdm,
                log_level="WARNING",
            )
        except NoCompatibleJobError as excp:
            print(str(excp))



"""
Helpers for the issue mutations
"""


from kili.authentication import KiliAuth

from ...queries.issue import QueriesIssue


def get_issue_number(auth: KiliAuth, project_id: str, type_: str):
    """Get the next available issue number

    Args:
        auth: Kili Auth
        project_id: Id of the project
        type_: type of the issue to add
    """
    kili = QueriesIssue(auth)
    issues = kili.issues(project_id=project_id, disable_tqdm=True, fields=["type", "issueNumber"])
    issue_number = (
        max(
            (issue["issueNumber"] for issue in issues if issue["type"] == type_),
            default=-1,
        )
        + 1
    )

    return issue_number



"""
Common code for the coco exporter.
"""

import json
import shutil
import time
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np
from PIL import Image
from typing_extensions import TypedDict

from kili.orm import Asset
from kili.services.export.format.base import AbstractExporter
from kili.services.types import Job, JobName, Jobs, ProjectId
from kili.utils.tempfile import TemporaryDirectory
from kili.utils.tqdm import tqdm

DATA_SUBDIR = "data"


# COCO format
class _CocoImage(TypedDict):
    id: int
    license: int
    file_name: str
    height: int
    width: int
    date_captured: None


class _CocoCategory(TypedDict):
    id: int
    name: str
    supercategory: str


class _CocoAnnotation(TypedDict):
    id: int
    image_id: int
    category_id: int
    bbox: List[int]
    segmentation: List[List[float]]  # [[x, y, x, y, x ...]]
    area: int
    iscrowd: int


class _CocoFormat(TypedDict):
    info: Dict  # type: ignore
    licenses: List[Dict]  # type: ignore
    categories: List[_CocoCategory]
    images: List[_CocoImage]
    annotations: List[_CocoAnnotation]


class CocoExporter(AbstractExporter):
    """
    Common code for Kili exporters.
    """

    download_media = True  # COCO is self-contained and requires paths to the media files

    def _check_arguments_compatibility(self):
        pass

    def process_and_save(self, assets: List[Asset], output_filename: Path):
        """
        Extract formatted annotations from labels.
        """
        clean_assets = self.process_assets(assets, self.label_format)

        with TemporaryDirectory() as tmp_dir:
            self._save_assets_export(
                clean_assets,
                tmp_dir,
            )
            self.create_readme_kili_file(tmp_dir)
            self.make_archive(tmp_dir, output_filename)

        self.logger.warning(output_filename)

    def _save_assets_export(self, assets: List[Asset], output_directory: Path):
        """
        Save the assets to a file and return the link to that file
        """
        jobs, title = self._get_project(self.kili, self.project_id)

        for job_name, job in jobs.items():
            if self._is_coco_compatible(job):
                _convert_kili_semantic_to_coco(
                    job_name=job_name,
                    assets=assets,
                    output_dir=Path(output_directory) / self.project_id,
                    job=job,
                    title=title,
                )
            else:
                self.logger.warning(f"Job {job_name} is not compatible with the COCO format.")

    @staticmethod
    def _get_project(kili, project_id: ProjectId) -> Tuple[Jobs, str]:
        projects = kili.projects(
            project_id=project_id, fields=["inputType", "jsonInterface", "title"], disable_tqdm=True
        )

        if len(projects) == 0:
            raise ValueError(
                "no such project. Maybe your KILI_API_KEY does not belong to a member of the"
                " project."
            )
        jobs = projects[0]["jsonInterface"].get("jobs", {})
        title = projects[0]["title"]
        return jobs, title

    @staticmethod
    def _is_coco_compatible(job: Job) -> bool:
        if "tools" not in job:
            return False
        return ("semantic" in job["tools"] or "rectangle" in job["tools"]) and job[
            "mlTask"
        ] == "OBJECT_DETECTION"


def _convert_kili_semantic_to_coco(
    job_name: JobName,
    assets: List[Asset],
    output_dir: Path,
    job: Job,
    title: str,
) -> Tuple[_CocoFormat, List[str]]:
    """
    creates the following structure on the disk:
    <dataset_dir>/
        data/
            <filename0>.<ext>
            <filename1>.<ext>
            ...
        labels.json


    We iterate on the assets and create a coco format for each asset.
    """
    infos_coco = {
        "year": time.strftime("%Y"),
        "version": "1.0",
        "description": f"{title} - Exported from Kili Python Client",
        "contributor": "Kili Technology",
        "url": "https://kili-technology.com",
        "date_created": datetime.now().isoformat(),
    }
    labels_json = _CocoFormat(
        info=infos_coco,
        licenses=[],
        categories=[],
        images=[],
        annotations=[],
    )

    # Prepare output folder
    data_dir = output_dir / DATA_SUBDIR
    data_dir.mkdir(parents=True, exist_ok=True)

    # Mapping category - category id
    cat_kili_id_to_coco_id = _get_kili_cat_id_to_coco_cat_id_mapping(job)
    labels_json["categories"] = _get_coco_categories(cat_kili_id_to_coco_id)
    labels_json["images"], labels_json["annotations"] = _get_coco_images_and_annotations(
        job_name, assets, data_dir, cat_kili_id_to_coco_id
    )

    label_file_name = output_dir / job_name / "labels.json"
    label_file_name.parent.mkdir(parents=True)
    with (output_dir / job_name / "labels.json").open("w") as outfile:
        json.dump(labels_json, outfile)

    classes = list(cat_kili_id_to_coco_id.keys())
    return labels_json, classes


def _get_kili_cat_id_to_coco_cat_id_mapping(job: Job) -> Dict[str, int]:
    cats = job["content"]["categories"]
    mapping_cat_name_cat_kili_id = {cat["name"]: catId for catId, cat in cats.items()}
    cat_kili_ids = list(mapping_cat_name_cat_kili_id.values())
    cat_kili_id_to_coco_id = {str(categoryId): i for i, categoryId in enumerate(cat_kili_ids)}
    return cat_kili_id_to_coco_id


def _get_coco_images_and_annotations(
    job_name, assets, data_dir, cat_kili_id_to_coco_id
) -> Tuple[List[_CocoImage], List[_CocoAnnotation]]:
    coco_images = []
    coco_annotations = []
    annotation_offset = -1
    for asset_i, asset in tqdm(
        enumerate(assets),
        desc="Convert to coco format",
    ):
        file_name = Path(shutil.copy(asset["content"], data_dir))
        width, height = Image.open(asset["content"]).size

        coco_image = _CocoImage(
            id=asset_i,
            license=0,
            file_name=str("data" / Path(file_name.parts[-1])),
            height=height,
            width=width,
            date_captured=None,
        )
        coco_images.append(coco_image)
        coco_img_annotations, annotation_offset = _get_coco_image_annotations(
            asset["latestLabel"]["jsonResponse"][job_name]["annotations"],
            cat_kili_id_to_coco_id,
            annotation_offset,
            coco_image,
        )
        coco_annotations.extend(coco_img_annotations)
    return coco_images, coco_annotations


def _get_coco_image_annotations(
    annotations_: List[Dict],
    cat_kili_id_to_coco_id: Dict[str, int],
    annotation_offset: int,
    asset: _CocoImage,
) -> Tuple[List[_CocoAnnotation], int]:
    coco_annotations = []

    annotation_j = annotation_offset

    for annotation in annotations_:  # we do not use enumerate as some annotations may be empty
        annotation_j += 1

        if not annotation:
            print("continue")
            continue
        bounding_poly = annotation["boundingPoly"]
        p_x = [float(v["x"]) * asset["width"] for v in bounding_poly[0]["normalizedVertices"]]
        p_y = [float(v["y"]) * asset["height"] for v in bounding_poly[0]["normalizedVertices"]]
        poly_ = [(float(x), float(y)) for x, y in zip(p_x, p_y)]
        if len(poly_) < 3:
            print("A polygon must contain more than 2 points. Skipping this polygon...")
            continue

        poly = [p for x in poly_ for p in x]

        categories = annotation["categories"]
        coco_annotations.append(
            _CocoAnnotation(
                id=annotation_j,
                image_id=asset["id"],
                category_id=cat_kili_id_to_coco_id[categories[0]["name"]],
                bbox=[int(np.min(p_x)), int(np.min(p_y)), int(np.max(p_x)), int(np.max(p_y))],
                # Objects have only one connected part.
                # But a type of object can appear several times on the same image.
                # The limitation of the single connected part comes from Kili.
                segmentation=[poly],
                area=asset["height"] * asset["width"],
                iscrowd=0,
            )
        )
    return coco_annotations, annotation_j


def _get_coco_categories(cat_kili_id_to_coco_id) -> List[_CocoCategory]:
    categories_coco: List[_CocoCategory] = []
    for cat_kili_id, cat_coco_id in cat_kili_id_to_coco_id.items():
        categories_coco.append(
            {
                "id": cat_coco_id,
                "name": cat_kili_id,
                "supercategory": "",
            }
        )

    return categories_coco



"""
Plugins related mutations
"""

GQL_CREATE_PLUGIN = """
mutation(
  $pluginName: String!
  ) {
  data: createPlugin(
    data: {
      pluginName: $pluginName
    }
  )
}
"""

GQL_CREATE_PLUGIN_RUNNER = """
mutation(
  $pluginName: String!
  ) {
  data: createPluginRunner(
    data: {
      pluginName: $pluginName
    }
  )
}
"""

GQL_GENERATE_UPDATE_URL = """
mutation(
  $pluginName: String!
  ) {
  data: generateUpdateUrl(
    data: {
      pluginName: $pluginName
    }
  )
}
"""

GQL_UPDATE_PLUGIN_RUNNER = """
mutation(
  $pluginName: String!
  ) {
  data: updatePluginRunner(
    data: {
      pluginName: $pluginName
    }
  )
}
"""

GQL_ACTIVATE_PLUGIN_ON_PROJECT = """
mutation(
  $pluginName: String!
  $projectId: ID!
  ) {
  data: activatePluginOnProject(
    where: {
      pluginName: $pluginName
      projectId: $projectId
    }
  )
}
"""

GQL_DEACTIVATE_PLUGIN_ON_PROJECT = """
mutation(
  $pluginName: String!
  $projectId: ID!
  ) {
  data: deactivatePluginOnProject(
    where: {
      pluginName: $pluginName
      projectId: $projectId
    }
  )
}
"""

GQL_DELETE_PLUGIN = """
mutation(
  $pluginName: String!
  ) {
  data: deletePlugin(
    where: {
      pluginName: $pluginName
    }
  )
}
"""



"""
Queries of user queries
"""


def gql_users(fragment):
    """
    Return the GraphQL users query
    """
    return f"""
query($where: UserWhere!, $first: PageSize!, $skip: Int!) {{
  data: users(where: $where, first: $first, skip: $skip) {{
    {fragment}
  }}
}}
"""


GQL_USERS_COUNT = """
query($where: UserWhere!) {
  data: countUsers(where: $where)
}
"""

GQL_ME = """
query Me {
  data: me {
    id
    email
  }
}
"""



"""Project mutations."""
from typing import Optional

from typeguard import typechecked

from kili.authentication import KiliAuth
from kili.services.plugins import (
    PluginUploader,
    activate_plugin,
    deactivate_plugin,
    delete_plugin,
)


class MutationsPlugins:
    """Set of Plugins mutations."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth: KiliAuth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @typechecked
    def upload_plugin(
        self,
        file_path: str,
        plugin_name: Optional[str] = None,
        verbose: bool = True,
    ):
        # pylint: disable=line-too-long
        """Uploads a plugin.

        Args:
            file_path : Path to your .py file
            plugin_name: name of your plugin, if not provided, it will be the name from your file
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.upload_plugin(file_path="./path/to/my/file.py")
        """

        return PluginUploader(self.auth, file_path, plugin_name, verbose).create_plugin()

    @typechecked
    def activate_plugin_on_project(
        self,
        plugin_name: str,
        project_id: str,
    ):
        # pylint: disable=line-too-long
        """Activates a plugin on a project.

        Args:
            plugin_name: Name of the plugin
            project_id: Identifier of the project
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.activate_plugin_on_project(plugin_name="my_plugin_name", project_id="my_project_id")
        """

        return activate_plugin(self.auth, plugin_name, project_id)

    @typechecked
    def deactivate_plugin_on_project(
        self,
        plugin_name: str,
        project_id: str,
    ):
        # pylint: disable=line-too-long
        """Activates a plugin on a project.

        Args:
            plugin_name: Name of the plugin
            project_id: Identifier of the project
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.deactivate_plugin_on_project(plugin_name="my_plugin_name", project_id="my_project_id")
        """

        return deactivate_plugin(self.auth, plugin_name, project_id)

    @typechecked
    def delete_plugin(
        self,
        plugin_name: str,
    ):
        # pylint: disable=line-too-long
        """Deletes a plugin.

        Args:
            plugin_name: Name of the plugin
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.delete_plugin(plugin_name="my_plugin_name")
        """

        return delete_plugin(self.auth, plugin_name)

    @typechecked
    def update_plugin(
        self,
        file_path: str,
        plugin_name: str,
        verbose: bool = True,
    ):
        """Update a plugin with new code.

        Args:
            file_path : Path to your .py file
            plugin_name: Name of the plugin
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.update_plugin(plugin_name="my_plugin_name")
        """

        return PluginUploader(self.auth, file_path, plugin_name, verbose).update_plugin()



"""
GraphQL Client
"""


import json
import os
import random
import string
import threading
import time
import urllib
from datetime import datetime
from enum import Enum
from typing import Dict

import websocket

from kili import __version__


class GraphQLClientName(Enum):  # pylint: disable=too-few-public-methods
    """GraphQL client name."""

    SDK = "python-sdk"
    CLI = "python-cli"


class GraphQLClient:
    """
    A simple GraphQL client
    """

    def __init__(
        self,
        endpoint,
        client_name: GraphQLClientName,
        session=None,
        verify=True,
    ):
        self.endpoint = endpoint
        self.client_name = client_name
        self.headername = None
        self.session = session
        self.token = None
        self.verify = verify

    def execute(self, query, variables=None) -> Dict:
        """
        Execute a query

        Args:
            query
            variables
        """
        return self._send(query, variables)

    def inject_token(self, token, headername="Authorization"):
        """Inject a token.

        Args:
            token
            headername:
        """
        self.token = token
        self.headername = headername

    def _send(self, query, variables) -> Dict:
        """
        Send the query

        Args:
            query
            variables
        """
        data = {"query": query, "variables": variables}
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "apollographql-client-name": self.client_name.value,
            "apollographql-client-version": __version__,
        }

        if self.token is not None:
            if self.headername:
                headers[self.headername] = f"{self.token}"
            else:
                raise ValueError("headername must be defined.")

        if self.session is not None:
            req = None
            try:
                try:
                    number_of_trials = int(os.getenv("KILI_SDK_TRIALS_NUMBER", "10"))
                except ValueError:
                    number_of_trials = 10
                for trial_number in range(number_of_trials):
                    self.session.verify = self.verify
                    req = self.session.post(
                        self.endpoint, json.dumps(data).encode("utf-8"), headers=headers
                    )
                    errors_in_response = "errors" in req.json()
                    if req.status_code == 200 and not errors_in_response:
                        break
                    if req.status_code == 401:
                        raise Exception("Invalid API KEY")
                    if trial_number == number_of_trials - 1 and errors_in_response:
                        break
                    time.sleep(1)
                return req.json()  # type:ignore X
            except Exception as exception:
                if req is not None:
                    raise Exception(req.content) from exception
                raise exception

        req = urllib.request.Request(  # type: ignore
            self.endpoint, json.dumps(data).encode("utf-8"), headers
        )
        try:
            with urllib.request.urlopen(req) as response:  # type:ignore
                str_json = response.read().decode("utf-8")
                return json.loads(str_json)
        except urllib.error.HTTPError as error:  # type:ignore
            print((error.read()))
            print("")
            raise error


GQL_WS_SUBPROTOCOL = "graphql-ws"


class SubscriptionGraphQLClient:
    """
    A simple GraphQL client that works over Websocket as the transport
    protocol, instead of HTTP.
    This follows the Apollo protocol.
    https://github.com/apollographql/subscriptions-transport-ws/blob/master/PROTOCOL.md
    """

    # pylint: disable=too-many-instance-attributes, too-many-arguments

    def __init__(self, url):
        self.ws_url = url
        self._id = None
        self._paused = False
        self._connect()
        self._subscription_running = False
        self._st_id = None
        self.failed_connection_attempts = 0

    def _connect(self):
        """
        Handles the connection
        """
        # pylint: disable=no-member
        self._conn = websocket.create_connection(
            self.ws_url, on_message=self._on_message, subprotocols=[GQL_WS_SUBPROTOCOL]
        )
        self._created_at = datetime.now()
        self._conn.on_message = self._on_message

    def _reconnect(self):
        """
        Handles the reconnection
        """
        self._connect()
        self._subscription_running = True
        dt_string = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
        print(f"{dt_string} reconnected")
        self.failed_connection_attempts = 0

    def _on_message(self, message):
        """
        Handles messages

        Args:
            message : the message
        """
        data = json.loads(message)
        # skip keepalive messages
        if data["type"] != "ka":
            print(message)

    def _conn_init(self, headers=None, authorization=None):
        """
        Initializes the websocket connection

        Args:
            headers : Headers are necessary for Kili API v1
            authorization : Headers are necessary for Kili API v2
        """
        payload = {
            "type": "connection_init",
            "payload": {"headers": headers, "Authorization": authorization},
        }
        self._conn.send(json.dumps(payload))
        self._conn.recv()

    def _start(self, payload):
        """
        Handles start

        Args:
            payload
        """
        _id = gen_id()
        frame = {"id": _id, "type": "start", "payload": payload}
        self._conn.send(json.dumps(frame))
        return _id

    def _stop(self, _id):
        """
        Handles stop

        Args:
        - _id: connection id
        """
        payload = {"id": _id, "type": "stop"}
        self._conn.send(json.dumps(payload))
        return self._conn.recv()

    def query(self, query, variables=None, headers=None):
        """
        Sends a query

        Args:
            query
            variables
            headers
        """
        self._conn_init(headers)
        payload = {"headers": headers, "query": query, "variables": variables}
        _id = self._start(payload)
        res = self._conn.recv()
        self._stop(_id)
        return res

    def prepare_subscribe(self, query, variables, headers, callback, authorization):
        """
        Prepares a subscription

        Args:
            query
            variables
            headers
            callback: function executed after the subscription
            authorization: authorization header
        """
        self._conn_init(headers, authorization)
        payload = {"headers": headers, "query": query, "variables": variables}
        _cc = self._on_message if not callback else callback
        _id = self._start(payload)
        self._id = _id
        return _cc, _id

    def subscribe(self, query, variables=None, headers=None, callback=None, authorization=None):
        """
        Subscribes

        Args:
            query
            variables
            headers
            callback: function executed after the subscription
            authorization: authorization header
        """
        _cc, _id = self.prepare_subscribe(query, variables, headers, callback, authorization)

        def subs(_cc, _id):
            max_reconnections = 10
            self._subscription_running = True
            while (
                self._subscription_running and self.failed_connection_attempts < max_reconnections
            ):
                try:
                    response = json.loads(self._conn.recv())
                    if response["type"] == "error" or response["type"] == "complete":
                        print(response)
                        self._stop_subscribe(_id)
                        break
                    if response["type"] != "ka" and not self._paused:
                        _cc(_id, response)  # type:ignore
                    time.sleep(1)
                except (
                    websocket._exceptions.WebSocketConnectionClosedException  # pylint: disable=no-member,protected-access
                ) as error:
                    self.failed_connection_attempts += 1
                    dt_string = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
                    error_message = str(error)
                    print(f"{dt_string} Connection closed error : {error_message}")
                    print(
                        "Will try to reconnect"
                        f" {max_reconnections - self.failed_connection_attempts} times..."
                    )
                    self._reconnect()
                    _cc, _id = self.prepare_subscribe(
                        query, variables, headers, callback, authorization
                    )
                    continue
            print(f"Did not reconnect successfully after {max_reconnections} attempts")

        self._st_id = threading.Thread(target=subs, args=(_cc, _id))
        self._st_id.start()
        return _id

    def _stop_subscribe(self, _id):
        self._subscription_running = False
        self._stop(_id)

    def close(self):
        """
        Handles close
        """
        self._conn.close()

    def pause(self):
        """
        Handles pause
        """
        self._paused = True

    def unpause(self):
        """
        Handles unpause
        """
        self._paused = False

    def get_lifetime(self):
        """
        Return the lifetime
        """
        return (datetime.now() - self._created_at).seconds

    def reset_timeout(self):
        """
        Resets the timeout
        """
        self._reconnect()


def gen_id(size=6, chars=string.ascii_letters + string.digits):
    """
    Generate random alphanumeric id

    Args:
        size: length of the id
        chars: chars used to generate the id
    """
    return "".join(random.choice(chars) for _ in range(size))



"""Common arguments and options for the CLI"""
from typing import List

import click

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
ROLES = ["ADMIN", "TEAM_MANAGER", "REVIEWER", "LABELER"]


class Options:  # pylint: disable=too-few-public-methods
    """Common options for the CLI"""

    api_key = click.option(
        "--api-key",
        type=str,
        default=None,
        help=("Your Kili API key. "),
        show_default=('"KILI_API_KEY" environment variable'),
    )

    endpoint = click.option(
        "--endpoint",
        type=str,
        default=None,
        help=("Kili API Endpoint. "),
        show_default=(
            '"KILI_API_ENDPOINT" environment variable,'
            ' Kili SAAS: "https://cloud.kili-technology.com/api/label/v2/graphql"'
        ),
    )

    project_id = click.option("--project-id", type=str, required=True, help="Id of the project")

    tablefmt = click.option(
        "--stdout-format",
        "tablefmt",
        type=str,
        default="plain",
        help="Defines how the output table is formatted "
        "(see https://pypi.org/project/tabulate/, default: plain).",
    )

    from_project = click.option(
        "--from-project",
        "project_id_src",
        type=str,
        help="project_id of another Kili project",
    )

    role = click.option(
        "--role",
        type=click.Choice(ROLES),
        default=None,
        show_default="LABELER",
        help="Project role of the added user(s).",
    )

    script_name = click.option(
        "--script-name",
        type=str,
        default=None,
        help=(
            "Script's name in Kili database."
            "If not provided, it will be the file name on your machine"
        ),
    )

    verbose = click.option(
        "--verbose", type=bool, is_flag=True, default=False, help="Show more logs"
    )


def from_csv(required_columns: List[str], optionnal_columns: List[str]):
    """--from-csv shared click option"""
    help_ = (
        "path to a csv file with required columns:"
        + ", ".join(required_columns)
        + " required columns: "
        + ", ".join(optionnal_columns)
    )

    return click.option(
        "--from-csv",
        "csv_path",
        type=click.Path(),
        help=help_,
    )


class Arguments:  # pylint: disable=too-few-public-methods
    """Common arguments for the CLI"""

    files = click.argument("files", type=click.Path(), required=False, nargs=-1)

    emails = click.argument("emails", type=str, required=False, nargs=-1)

    project_id = click.argument("project_id", type=str, required=True)



"""CLI's project label subcommand"""


import os
from typing import Optional, Tuple

import click
from typing_extensions import get_args

from kili.cli.common_args import Arguments, Options, from_csv
from kili.cli.helpers import get_kili_client
from kili.services import label_import
from kili.services.label_import.types import LabelFormat


def type_check_label(key, value):
    """type check value based on key"""
    if key == "json_response_path" and not (os.path.isfile(value) and value.endswith(".json")):
        return f"{value} is not a valid path to a json file, "

    return ""


@click.command(name="label")
@Options.api_key
@Options.endpoint
@Arguments.files
@from_csv(["external_id", "json_response_path"], [])
@Options.project_id
@click.option(
    "--prediction",
    "is_prediction",
    type=bool,
    is_flag=True,
    default=False,
    help=(
        "Tells to import labels as predictions, which means that they will appear "
        "as pre-annotations in the Kili interface"
    ),
)
@click.option(
    "--model-name",
    type=str,
    help="Name of the model that generated predictions, if labels are sent as predictions",
)
@Options.verbose
@click.option(
    "--input-format",
    type=click.Choice(get_args(LabelFormat)),
    help="Format in which the labels are encoded",
    default="kili",
    show_default='"raw" kili format',
)
@click.option(
    "--metadata-file",
    type=str,
    help="File containing format metadata (if relevant to the input format)",
)
@click.option(
    "--target-job",
    type=str,
    help="Job name in the project where to upload the labels (if relevant to the input format)",
)
# pylint: disable=too-many-arguments, too-many-locals
def import_labels(
    api_key: Optional[str],
    endpoint: Optional[str],
    files: Optional[Tuple[str, ...]],
    csv_path: str,
    project_id: str,
    is_prediction: bool,
    model_name: Optional[str],
    verbose: bool,
    input_format: str,
    metadata_file: Optional[str],
    target_job: Optional[str],
):
    """
    Import labels or predictions

    Files can be paths to files or to folders. You can provide several paths separated by spaces.
    The labels to import have to be in the Kili format and stored in a json file.
    File's name must be equal to asset's external_id.

    If no files are provided, --from-csv can be used to import
    assets from a CSV file with two columns:

    \b
      - `label_asset_external_id`: external id for which you want to import labels.
      - `label_asset_id`: asset id for which you want to import labels (mutual exclusive with the
    field above, and not available for predictions)
      - `path`: paths to the json files containing the json_response to upload.

    Additional columns can be provided in the CSV file, see `.append_labels` in the Python client
    documentation:

    \b
      - label_type
      - seconds_to_label
      - author_id

    \b
    !!! Examples "CSV file template for the raw Kili format"
        ```
        label_asset_external_id,path
        asset1,./labels/label_asset1.json
        asset2,./labels/label_asset2.json
        ```

     \b
    !!! Examples "CSV file template for a Yolo format"
        ```
        label_asset_external_id,path
        asset1,./labels/label_asset1.txt
        asset2,./labels/label_asset2.txt
        ```

    \b
    !!! Examples
        To import default labels:
        ```
        kili project label \\
             dir1/dir2/ dir1/dir3/test1.json \\
            --project-id <project_id>
        ```
        ```
        kili project label \\
            --from-csv path/to/file.csv \\
            --project-id <project_id>
        ```
        To import labels as predictions:
        ```
        kili project label \\
            --from-csv path/to/file.csv \\
            --project-id <project_id> \\
            --prediction \\
            --model-name YOLO-run-3
        ```
        To import labels as predictions in the Yolo v5 format into a target job:
        ```
        kili project label \\
            --from-csv path/to/file.csv \\
            --project-id <project_id> \\
            --prediction \\
            --model-name YOLO-v5 \\
            --metadata-file classes.yml \\
            --target-job IMAGE_DETECTION_JOB \\
            --input-format yolo_v5
        ```


    """
    if is_prediction and model_name is None:
        raise ValueError(
            "When importing labels as prediction, "
            "you must provide a model name with the --model-name option"
        )

    kili = get_kili_client(api_key=api_key, api_endpoint=endpoint)

    label_import.import_labels_from_files(
        kili,
        csv_path,
        list(files or []),
        metadata_file,
        project_id,
        input_format,
        target_job,
        not verbose,
        "INFO" if verbose else "WARNING",
        model_name,
        is_prediction,
    )



"""
Label mutations
"""

import warnings
from itertools import repeat
from json import dumps
from typing import Dict, List, Optional

from typeguard import typechecked

from kili.enums import LabelType
from kili.helpers import deprecate, format_result, infer_id_from_external_id
from kili.mutations.label.queries import (
    GQL_APPEND_TO_LABELS,
    GQL_CREATE_HONEYPOT,
    GQL_CREATE_PREDICTIONS,
    GQL_UPDATE_PROPERTIES_IN_LABEL,
)
from kili.orm import Label
from kili.services.label_import import import_labels_from_dict
from kili.utils.pagination import _mutate_from_paginated_call


class MutationsLabel:
    """Set of Label mutations."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initializes the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @typechecked
    def create_predictions(
        self,
        project_id: str,
        external_id_array: List[str],
        model_name_array: List[str],
        json_response_array: List[dict],
    ) -> Label:
        # pylint: disable=line-too-long
        """Create predictions for specific assets.

        Args:
            project_id: Identifier of the project
            external_id_array: The external identifiers of the assets for which we want to add predictions
            model_name_array: In case you want to precise from which model the label originated
            json_response_array: The predictions are given here. For examples,
                see [the recipe](https://docs.kili-technology.com/recipes/importing-labels-and-predictions).

        Returns:
            A result object which indicates if the mutation was successful, or an error message.

        !!! example "Recipe"
            For more detailed examples on how to create predictions, see [the recipe](https://docs.kili-technology.com/recipes/importing-labels-and-predictions).
        """
        assert len(external_id_array) == len(
            json_response_array
        ), "IDs list and predictions list should have the same length"
        assert len(external_id_array) == len(
            model_name_array
        ), "IDs list and model names list should have the same length"
        if len(external_id_array) == 0:
            warnings.warn("Empty IDs and prediction list")

        properties_to_batch = {
            "external_id_array": external_id_array,
            "model_name_array": model_name_array,
            "json_response_array": json_response_array,
        }

        def generate_variables(batch):
            return {
                "data": {
                    "modelNameArray": batch["model_name_array"],
                    "jsonResponseArray": [dumps(elem) for elem in batch["json_response_array"]],
                },
                "where": {
                    "externalIdStrictlyIn": batch["external_id_array"],
                    "project": {"id": project_id},
                },
            }

        results = _mutate_from_paginated_call(
            self,
            properties_to_batch,
            generate_variables,
            GQL_CREATE_PREDICTIONS,
        )
        return format_result("data", results[0], Label)

    @deprecate(
        msg=(
            "append_to_labels method is deprecated. Please use append_labels instead. This new"
            " function allows to import several labels 10 times faster."
        )
    )
    @typechecked
    def append_to_labels(
        self,
        json_response: dict,
        author_id: Optional[str] = None,
        label_asset_external_id: Optional[str] = None,
        label_asset_id: Optional[str] = None,
        label_type: str = "DEFAULT",
        project_id: Optional[str] = None,
        seconds_to_label: Optional[int] = 0,
    ):
        """
        !!! danger "[DEPRECATED]"
            append_to_labels method is deprecated. Please use append_labels instead.
            This new function allows to import several labels 10 times faster.

        Append a label to an asset.

        Args:
            json_response: Label is given here
            author_id: ID of the author of the label
            label_asset_external_id: External identifier of the asset
            label_asset_id: Identifier of the asset
            project_id: Identifier of the project
            label_type: Can be one of `DEFAULT`, `PREDICTION`, `REVIEW` or `INFERENCE`
            seconds_to_label: Time to create the label

        !!! warning
            Either provide `label_asset_id` or `label_asset_external_id` and `project_id`

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.append_to_labels(label_asset_id=asset_id, json_response={...})

        """
        if author_id is None:
            author_id = self.auth.user_id
        label_asset_id = infer_id_from_external_id(
            self, label_asset_id, label_asset_external_id, project_id
        )
        variables = {
            "data": {
                "authorID": author_id,
                "jsonResponse": dumps(json_response),
                "labelType": label_type,
                "secondsToLabel": seconds_to_label,
            },
            "where": {"id": label_asset_id},
        }
        result = self.auth.client.execute(GQL_APPEND_TO_LABELS, variables)
        return format_result("data", result, Label)

    @typechecked
    def append_labels(
        self,
        asset_id_array: List[str],
        json_response_array: List[Dict],
        author_id_array: Optional[List[str]] = None,
        seconds_to_label_array: Optional[List[int]] = None,
        model_name: Optional[str] = None,
        label_type: LabelType = "DEFAULT",
    ) -> List:
        """Append labels to assets.

        Args:
            asset_id_array: list of asset ids to append labels on
            json_response_array: list of labels to append
            author_id_array: list of the author id of the labels
            seconds_to_label_array: list of times taken to produce the label, in seconds
            model_name: Only useful when uploading predictions.
                Name of the model when uploading predictions
            label_type: Can be one of `AUTOSAVE`, `DEFAULT`, `PREDICTION`, `REVIEW` or `INFERENCE`

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.append_to_labels(
                    asset_id_array=['cl9wmlkuc00050qsz6ut39g8h', 'cl9wmlkuw00080qsz2kqh8aiy'],
                    json_response_array=[{...}, {...}]
                )
        """
        if label_type == "PREDICTION" and model_name is None:
            raise ValueError("You must provide model_name when uploading predictions")
        if len(asset_id_array) != len(json_response_array):
            raise ValueError("asset_id_array and json_response_array should have the same length")
        for array in [seconds_to_label_array, author_id_array]:
            if array is not None and len(array) != len(asset_id_array):
                raise ValueError("All arrays should have the same length")
        labels = [
            {
                "asset_id": asset_id,
                "json_response": json_response,
                "seconds_to_label": seconds_to_label,
                "model_name": model_name,
                "author_id": author_id,
            }
            for (asset_id, json_response, seconds_to_label, author_id, model_name) in list(
                zip(
                    asset_id_array,
                    json_response_array,
                    seconds_to_label_array or [None] * len(asset_id_array),
                    author_id_array or [None] * len(asset_id_array),
                    repeat(model_name),
                )
            )
        ]
        return import_labels_from_dict(self, labels, label_type)

    @typechecked
    def update_properties_in_label(
        self,
        label_id: str,
        seconds_to_label: Optional[int] = None,
        model_name: Optional[str] = None,
        json_response: Optional[dict] = None,
    ) -> Label:
        """Update properties of a label.

        Args:
            label_id: Identifier of the label
            seconds_to_label: Time to create the label
            model_name: Name of the model
            json_response: The label is given here

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.update_properties_in_label(label_id=label_id, json_response={...})
        """
        formatted_json_response = None if json_response is None else dumps(json_response)
        variables = {
            "labelID": label_id,
            "secondsToLabel": seconds_to_label,
            "modelName": model_name,
            "jsonResponse": formatted_json_response,
        }
        result = self.auth.client.execute(GQL_UPDATE_PROPERTIES_IN_LABEL, variables)
        return format_result("data", result, Label)

    @typechecked
    def create_honeypot(
        self,
        json_response: dict,
        asset_external_id: Optional[str] = None,
        asset_id: Optional[str] = None,
        project_id: Optional[str] = None,
    ) -> Label:
        """Create honeypot for an asset.

        !!! info
            Uses the given `json_response` to create a `REVIEW` label.
            This enables Kili to compute a`honeypotMark`,
            which measures the similarity between this label and other labels.

        Args:
            json_response: The JSON response of the honeypot label of the asset
            asset_id: Identifier of the asset
                Either provide asset_id or asset_external_id and project_id
            asset_external_id: External identifier of the asset
                Either provide asset_id or asset_external_id and project_id
            project_id: Identifier of the project
                Either provide asset_id or asset_external_id and project_id

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        asset_id = infer_id_from_external_id(self, asset_id, asset_external_id, project_id)

        variables = {
            "data": {"jsonResponse": dumps(json_response)},
            "where": {"id": asset_id},
        }
        result = self.auth.client.execute(GQL_CREATE_HONEYPOT, variables)
        return format_result("data", result, Label)



"""
Issue mutations
"""

from typing import Dict, List, Optional, cast

from typeguard import typechecked
from typing_extensions import Literal

from ...helpers import format_result
from ...queries.label import QueriesLabel
from .helpers import get_issue_number
from .queries import GQL_APPEND_TO_ISSUES


class MutationsIssue:
    """Set of Issue mutations."""

    # pylint: disable=too-few-public-methods,too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @typechecked
    def append_to_issues(
        self,
        label_id: str,
        project_id: str,
        object_mid: Optional[str] = None,
        text: Optional[str] = None,
        type_: Literal["ISSUE", "QUESTION"] = "ISSUE",
    ) -> Dict:
        """Create an issue.

        Args:
            label_id: Id of the label to add an issue to
            object_mid: Mid of the object in the label to associate the issue to
            type_: type of the issue to add. Can be either "ISSUE" or "QUESTION"
            text: If given, write a comment related to the issue
            project_id: Id of the project

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        issue_number = get_issue_number(self.auth, project_id, type_)
        try:
            asset_id = cast(
                List[Dict],
                list(
                    QueriesLabel(self.auth).labels(
                        project_id=project_id,
                        label_id=label_id,
                        fields=["labelOf.id"],
                        disable_tqdm=True,
                    )
                )[0]["labelOf"]["id"],
            )
        except:
            # pylint: disable=raise-missing-from
            raise ValueError(
                f"Label ID {label_id} does not exist in the project of ID {project_id}"
            )
        variables = {
            "data": {
                "issueNumber": issue_number,
                "labelID": label_id,
                "objectMid": object_mid,
                "type": type_,
            },
            "where": {"id": asset_id},
        }
        if text:
            variables["data"]["text"] = text

        result = self.auth.client.execute(GQL_APPEND_TO_ISSUES, variables)
        return format_result("data", result)



"""Module for managing bucket's signed urls"""


import itertools
from typing import List, Union
from urllib.parse import parse_qs, urlparse

import cuid
import requests
from tenacity import retry
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_random

from kili.authentication import KiliAuth
from kili.graphql.operations.asset.queries import GQL_CREATE_UPLOAD_BUCKET_SIGNED_URLS

AZURE_STRING = "blob.core.windows.net"
GCP_STRING = "storage.googleapis.com"
GCP_STRING_PUBLIC = "storage.cloud.google.com"

MAX_NUMBER_SIGNED_URLS_TO_FETCH = 30


def generate_unique_id():
    """
    Generates a unique id
    """
    return cuid.cuid()


def request_signed_urls(auth: KiliAuth, file_urls: List[str]):
    """
    Get upload signed URLs
    Args:
        auth: Kili Auth
        file_paths: the paths in Kili bucket of the data you upload. It must respect
            the convention projects/:projectId/assets/xxx.
    """
    size = len(file_urls)
    file_batches = [
        file_urls[i : i + MAX_NUMBER_SIGNED_URLS_TO_FETCH]
        for i in range(0, size, MAX_NUMBER_SIGNED_URLS_TO_FETCH)
    ]

    def get_file_batch_urls(file_paths: List[str]) -> List[str]:
        payload = {
            "filePaths": file_paths,
        }
        urls_response = auth.client.execute(GQL_CREATE_UPLOAD_BUCKET_SIGNED_URLS, payload)
        return urls_response["data"]["urls"]

    return [*itertools.chain(*map(get_file_batch_urls, file_batches))]


@retry(stop=stop_after_attempt(3), wait=wait_random(min=1, max=2))
def upload_data_via_rest(url_with_id: str, data: Union[str, bytes], content_type: str):
    """upload data in buckets' signed URL via REST
    Args:
        signed_urls: Bucket signed URLs to upload local files to
        path_array: a list of file paths, json or text to upload
        content_type: mimetype of the data
    """
    if content_type == "text/plain":
        content_type += "; charset=utf-8"
    headers = {"Content-type": content_type}
    url_to_use_for_upload = url_with_id.split("&id=")[0]
    if "blob.core.windows.net" in url_to_use_for_upload:
        headers["x-ms-blob-type"] = "BlockBlob"

    response = requests.put(url_to_use_for_upload, data=data, headers=headers, timeout=30)
    response.raise_for_status()
    return url_with_id


def clean_signed_url(url: str, endpoint: str):
    """
    return a cleaned signed url for frame upload
    """
    query = urlparse(url).query
    id_param = parse_qs(query)["id"][0]
    base_path = endpoint.replace("/graphql", "/files").replace("http://", "https://")
    return f"{base_path}?id={id_param}"



"""
Base class for all formatters and other utility classes.
"""

import logging
import shutil
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional, cast

from kili.orm import Asset, JobMLTask, JobTool, Label
from kili.services.export.repository import AbstractContentRepository
from kili.services.export.tools import fetch_assets
from kili.services.export.types import ExportType, LabelFormat, SplitOption
from kili.services.types import ProjectId


class ExportParams(NamedTuple):
    """
    Contains all parameters that change the result of the export
    """

    assets_ids: Optional[List[str]]
    export_type: ExportType
    project_id: ProjectId
    label_format: LabelFormat
    split_option: SplitOption
    single_file: bool
    output_file: Path


class AbstractExporter(ABC):  # pylint: disable=too-many-instance-attributes

    """
    Abstract class defining the interface for all exporters.
    """

    download_media = False  # Whether or not we need to download the media for the given format.

    def __init__(
        self,
        export_params: ExportParams,
        kili,
        logger: logging.Logger,
        disable_tqdm: bool,
        content_repository: AbstractContentRepository,
    ):  # pylint: disable=too-many-arguments
        self.project_id: ProjectId = export_params.project_id
        self.assets_ids: Optional[List[str]] = export_params.assets_ids
        self.export_type: ExportType = export_params.export_type
        self.label_format: LabelFormat = export_params.label_format
        self.single_file: bool = export_params.single_file
        self.split_option: SplitOption = export_params.split_option
        self.disable_tqdm: bool = disable_tqdm
        self.kili = kili
        self.logger: logging.Logger = logger
        self.content_repository: AbstractContentRepository = content_repository
        self.output_file = export_params.output_file

    @abstractmethod
    def _check_arguments_compatibility(self):
        """
        Checks if the format can accept the
        """

    @abstractmethod
    def process_and_save(self, assets: List[Dict], output_filename: Path) -> None:
        """
        Converts the asset and save them into an archive file.
        """

    def make_archive(self, root_folder: Path, output_filename: Path) -> Path:
        """
        Make the export archive
        """
        path_folder = root_folder / self.project_id
        path_archive = shutil.make_archive(str(path_folder), "zip", path_folder)
        shutil.copy(path_archive, output_filename)
        return output_filename

    def get_project_and_init(self):
        """
        Get and validate the project
        """
        json_interface = list(
            self.kili.projects(
                project_id=self.project_id, fields=["jsonInterface"], disable_tqdm=True
            )
        )[0]["jsonInterface"]

        ml_task = JobMLTask.ObjectDetection
        tool = JobTool.Rectangle

        return json_interface, ml_task, tool

    def create_readme_kili_file(self, root_folder: Path) -> None:
        """
        Create a README.kili.txt file to give information about exported labels
        """
        readme_file_name = root_folder / self.project_id / "README.kili.txt"
        project_info = list(
            self.kili.projects(
                project_id=self.project_id, fields=["title", "id", "description"], disable_tqdm=True
            )
        )[0]
        readme_file_name.parent.mkdir(parents=True, exist_ok=True)
        with readme_file_name.open("wb") as fout:
            fout.write("Exported Labels from KILI\n=========================\n\n".encode())
            fout.write(f"- Project name: {project_info['title']}\n".encode())
            fout.write(f"- Project identifier: {self.project_id}\n".encode())
            fout.write(f"- Project description: {project_info['description']}\n".encode())
            fout.write(f'- Export date: {datetime.now().strftime("%Y%m%d-%H%M%S")}\n'.encode())
            fout.write(f"- Exported format: {self.label_format}\n".encode())
            fout.write(f"- Exported labels: {self.export_type}\n".encode())

    def export_project(
        self,
    ) -> None:
        """
        Export a project to a json.
        Return the name of the exported archive file in the bucket.
        """
        self._check_arguments_compatibility()
        self.logger.warning("Fetching assets...")
        assets = fetch_assets(
            self.kili,
            project_id=self.project_id,
            asset_ids=self.assets_ids,
            export_type=self.export_type,
            label_type_in=["DEFAULT", "REVIEW"],
            disable_tqdm=self.disable_tqdm,
            download_media=self.download_media,
        )
        self.process_and_save(assets, self.output_file)

    @staticmethod
    def _filter_out_autosave_labels(assets: List[Asset]) -> List[Asset]:
        """
        Removes AUTOSAVE labels from exports
        """
        clean_assets = []
        for asset in assets:
            labels = asset.get("labels", [])
            clean_labels = list(filter(lambda label: label["labelType"] != "AUTOSAVE", labels))
            if clean_labels:
                asset["labels"] = clean_labels
            clean_assets.append(asset)
        return clean_assets

    @staticmethod
    def _format_json_response(label: Label, label_format: LabelFormat):
        """
        Format the label JSON response in the requested format
        """
        formatted_json_response = label.json_response(_format=label_format)
        json_response = {}
        for key, value in cast(Dict, formatted_json_response).items():
            if key.isdigit():
                json_response[int(key)] = value
                continue
            json_response[key] = value
        label["jsonResponse"] = json_response
        return label

    @staticmethod
    def process_assets(assets: List[Asset], label_format: LabelFormat) -> List[Asset]:
        """
        Format labels in the requested format, and filter out autosave labels
        """
        assets_in_format = []
        for asset in assets:
            if "labels" in asset:
                labels_of_asset = []
                for label in asset["labels"]:
                    clean_label = AbstractExporter._format_json_response(label, label_format)
                    labels_of_asset.append(clean_label)
                asset["labels"] = labels_of_asset
            if "latestLabel" in asset:
                label = asset["latestLabel"]
                if label is not None:
                    clean_label = AbstractExporter._format_json_response(label, label_format)
                    asset["latestLabel"] = clean_label
            assets_in_format.append(asset)

        clean_assets = AbstractExporter._filter_out_autosave_labels(assets_in_format)
        return clean_assets



"""
Import service exceptions
"""


class MimeTypeError(Exception):
    """
    Raised when the mime type of a file is not found or
    is not compatible with the project type to import in
    """


class ImportValidationError(Exception):
    """
    Raised when data given to import does not follow a right format
    """



"""
Common code for the yolo exporter.
"""

import json
from pathlib import Path
from typing import List

from kili.orm import Asset
from kili.services.export.format.base import AbstractExporter
from kili.utils.tempfile import TemporaryDirectory


class KiliExporter(AbstractExporter):
    """
    Common code for Kili exporters.
    """

    def _check_arguments_compatibility(self):
        """
        Checks if the format can accept the export options
        """

    def _save_assets_export(self, assets: List[Asset], output_filename: Path):
        """
        Save the assets to a file and return the link to that file
        """
        self.logger.info("Exporting to kili format...")

        with TemporaryDirectory() as root_folder:
            base_folder = root_folder / self.project_id
            base_folder.mkdir(parents=True)
            if self.single_file:
                project_json = json.dumps(assets, sort_keys=True, indent=4)
                with (base_folder / "data.json").open("wb") as output_file:
                    output_file.write(project_json.encode("utf-8"))
            else:
                labels_folder = base_folder / "labels"
                labels_folder.mkdir(parents=True)
                for asset in assets:
                    external_id = asset["externalId"].replace(" ", "_")
                    asset_json = json.dumps(asset, sort_keys=True, indent=4)
                    with (labels_folder / f"{external_id}.json").open("wb") as output_file:
                        output_file.write(asset_json.encode("utf-8"))
            self.create_readme_kili_file(root_folder)
            self.make_archive(root_folder, output_filename)

        self.logger.warning(output_filename)

    def process_and_save(self, assets: List[Asset], output_filename: Path):
        """
        Extract formatted annotations from labels and save the json in the buckets.
        """
        clean_assets = self.process_assets(assets, self.label_format)
        return self._save_assets_export(
            clean_assets,
            output_filename,
        )



"""Kili CLI"""
import click

from kili import __version__
from kili.cli.common_args import CONTEXT_SETTINGS
from kili.cli.project import project


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(__version__)
def cli():
    """Kili Command line Interface

    To get all the available commands, please type: `kili project --help`.
    """


cli.add_command(project, name="project")


def main() -> None:
    """Execute the main function of the command line."""
    cli()


if __name__ == "__main__":
    main()
